# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Pixmap widget"""

import os.path
from utils.pixmapcache import getIcon
from .qt import (QPalette, QSizePolicy, QScrollArea, QImage, QPixmap, QAction,
                 QLabel, QToolBar, QWidget, QHBoxLayout, QApplication, QMenu,
                 QCursor, QShortcut, Qt, QSize, pyqtSignal)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase
from .outsidechanges import OutsideChangeWidget


FORMAT_STRINGS = {
    QImage.Format_Invalid: "invalid",
    QImage.Format_Mono: "1-bit per pixel",
    QImage.Format_MonoLSB: "1-bit per pixel",
    QImage.Format_Indexed8: "8-bit indexes",
    QImage.Format_RGB32: "32-bit RG",
    QImage.Format_ARGB32: "32-bit ARGB",
    QImage.Format_ARGB32_Premultiplied: "32-bit ARGB",
    QImage.Format_RGB16: "16-bit RGB",
    QImage.Format_ARGB8565_Premultiplied: "24-bit ARGB",
    QImage.Format_RGB666: "24-bit RGB",
    QImage.Format_ARGB6666_Premultiplied: "24-bit ARGB",
    QImage.Format_RGB555: "16-bit RGB",
    QImage.Format_ARGB8555_Premultiplied: "24-bit ARGB",
    QImage.Format_RGB888: "24-bit RGB",
    QImage.Format_RGB444: "16-bit RGB",
    QImage.Format_ARGB4444_Premultiplied: "16-bit ARGB"}


class PixmapWidget(QScrollArea):

    """The pixmap widget"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        QScrollArea.__init__(self, parent)

        self.pixmapLabel = QLabel()
        self.pixmapLabel.setBackgroundRole(QPalette.Base)
        self.pixmapLabel.setSizePolicy(QSizePolicy.Ignored,
                                       QSizePolicy.Ignored)
        self.pixmapLabel.setScaledContents(True)

        self.zoom = 1.0
        self.info = ""
        self.formatInfo = ""
        self.fileSize = 0

        self.setBackgroundRole(QPalette.Dark)
        self.setWidget(self.pixmapLabel)
        self.setAlignment(Qt.AlignCenter)

    def loadFromFile(self, fileName):
        """Loads a pixmap from a file"""
        image = QImage(fileName)
        if image.isNull():
            raise Exception("Unsupported pixmap format (" + fileName + ")")

        self.pixmapLabel.setPixmap(QPixmap.fromImage(image))
        self.pixmapLabel.adjustSize()

        self.fileSize = os.path.getsize(fileName)
        if self.fileSize < 1024:
            fileSizeString = str(self.fileSize) + "bytes"
        else:
            kiloBytes = self.fileSize / 1024
            if (self.fileSize % 1024) >= 512:
                kiloBytes += 1
            fileSizeString = str(kiloBytes) + "kb"
        self.info = str(image.width()) + "px/" + \
            str(image.height()) + "px/" + fileSizeString
        try:
            self.formatInfo = FORMAT_STRINGS[image.format()]
        except:
            self.formatInfo = "Unknown"

    def setPixmap(self, pixmap):
        """Shows the provided pixmap"""
        pix = QPixmap.fromImage(pixmap)
        self.pixmapLabel.setPixmap(pix)
        self.pixmapLabel.adjustSize()

        self.info = str(pix.width()) + "px/" + str(pix.height()) + "px"
        self.formatInfo = str(pix.depth()) + " bpp"

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            self.sigEscapePressed.emit()
            event.accept()
        else:
            QScrollArea.keyPressEvent(self, event)

    def resetZoom(self):
        """Resets the zoom"""
        self.zoom = 1.0
        self.pixmapLabel.adjustSize()

    def doZoom(self, factor):
        """Performs zooming"""
        self.zoom *= factor
        self.pixmapLabel.resize(self.zoom * self.pixmapLabel.pixmap().size())

        self.__adjustScrollBar(self.horizontalScrollBar(), factor)
        self.__adjustScrollBar(self.verticalScrollBar(), factor)

    def __adjustScrollBar(self, scrollBar, factor):
        """Adjusts a scrollbar by a certain factor"""
        scrollBar.setValue(int(factor * scrollBar.value() +
                               ((factor - 1) * scrollBar.pageStep() / 2)))

    def setReadOnly(self, newValue):
        """Make it similar to a text editor"""


class PixmapTabWidget(QWidget, MainWindowTabWidgetBase):

    """Pixmap viewer tab widget"""

    sigEscapePressed = pyqtSignal()
    reloadRequst = pyqtSignal()
    reloadAllNonModifiedRequest = pyqtSignal()

    def __init__(self, parent):
        MainWindowTabWidgetBase.__init__(self)
        QWidget.__init__(self, parent)

        self.__editorsManager = parent
        self.__viewer = PixmapWidget()
        self.__fileName = ""
        self.__shortName = ""

        self.__viewer.sigEscapePressed.connect(self.__onEsc)
        self.__viewer.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__viewer.customContextMenuRequested.connect(self.__onContextMenu)

        self.__createLayout()

        self.__diskModTime = None
        self.__diskSize = None
        self.__reloadDlgShown = False

        self.__vcsStatus = None

    def __createLayout(self):
        """Creates the toolbar and layout"""
        # Buttons
        printButton = QAction(getIcon('printer.png'), 'Print', self)
        # printButton.setShortcut('Ctrl+')
        printButton.triggered.connect(self.__onPrint)
        printButton.setVisible(False)

        printPreviewButton = QAction(getIcon('printpreview.png'),
                                     'Print preview', self)
        # printPreviewButton.setShortcut('Ctrl+')
        printPreviewButton.triggered.connect(self.__onPrintPreview)
        printPreviewButton.setVisible(False)

        fixedSpacer = QWidget()
        fixedSpacer.setFixedHeight(16)

        zoomInButton = QAction(getIcon('zoomin.png'), 'Zoom in (Ctrl+=)', self)
        zoomInButton.setShortcut('Ctrl+=')
        zoomInButton.triggered.connect(self.onZoomIn)
        self.__zoomInSynonim = QShortcut("Ctrl++", self)
        self.__zoomInSynonim.activated.connect(self.onZoomIn)

        zoomOutButton = QAction(getIcon('zoomout.png'),
                                'Zoom out (Ctrl+-)', self)
        zoomOutButton.setShortcut('Ctrl+-')
        zoomOutButton.triggered.connect(self.onZoomOut)

        zoomResetButton = QAction(getIcon('zoomreset.png'),
                                  'Zoom reset (Ctrl+0)', self)
        zoomResetButton.setShortcut('Ctrl+0')
        zoomResetButton.triggered.connect(self.onZoomReset)

        # Toolbar
        toolbar = QToolBar(self)
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.RightToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)
        toolbar.addAction(printPreviewButton)
        toolbar.addAction(printButton)
        toolbar.addWidget(fixedSpacer)
        toolbar.addAction(zoomInButton)
        toolbar.addAction(zoomOutButton)
        toolbar.addAction(zoomResetButton)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)
        hLayout.addWidget(self.__viewer)
        hLayout.addWidget(toolbar)

        self.__outsideChangesBar = OutsideChangeWidget(self.__viewer)
        self.__outsideChangesBar.sigReloadRequest.connect(self.__onReload)
        self.__outsideChangesBar.reloadAllNonModifiedRequest.connect(
            self.reloadAllNonModified)
        self.__outsideChangesBar.hide()

        self.setLayout(hLayout)

    def setFocus(self):
        """Overridden setFocus"""
        if self.__outsideChangesBar.isHidden():
            self.__viewer.setFocus()
        else:
            self.__outsideChangesBar.setFocus()

    def loadFromFile(self, path):
        """Loads the content from the given file"""
        self.__viewer.loadFromFile(path)
        self.setFileName(os.path.abspath(path))

        # Memorize the modification date
        path = os.path.realpath(path)
        self.__diskModTime = os.path.getmtime(path)
        self.__diskSize = os.path.getsize(path)

    def setPixmap(self, pixmap):
        """Loads the provided pixmap"""
        self.__viewer.setPixmap(pixmap)
        self.__diskModTime = None
        self.__diskSize = None

    def __onPrint(self):
        """Triggered on the 'print' button"""

    def __onPrintPreview(self):
        """Triggered on the 'print preview' button"""

    def onZoomIn(self):
        """Triggered on the 'zoom in' button"""
        self.__viewer.doZoom(1.25)
        QApplication.processEvents()
        self.resizeBars()

    def onZoomOut(self):
        """Triggered on the 'zoom out' button"""
        self.__viewer.doZoom(0.8)
        QApplication.processEvents()
        self.resizeBars()

    def wheelEvent(self, event):
        """Mouse wheel event"""
        if QApplication.keyboardModifiers() == Qt.ControlModifier:
            angleDelta = event.angleDelta()
            if not angleDelta.isNull():
                if angleDelta.y() > 0:
                    self.onZoomIn()
                else:
                    self.onZoomOut()
            event.accept()
        else:
            QWidget.wheelEvent(self, event)

    def __onContextMenu(self, pos):
        """Triggered when a context menu is requested"""
        del pos     # unused argument
        pluginMenus = self.__editorsManager.getPluginMenus()
        if pluginMenus:
            contextMenu = QMenu()
            for pluginPath, pluginMenu in pluginMenus.iteritems():
                del pluginPath  # unused variable
                contextMenu.addMenu(pluginMenu)
            contextMenu.exec_(QCursor.pos())
            del contextMenu

    def onZoomReset(self):
        """Triggered on the 'zoom reset' button"""
        self.__viewer.resetZoom()
        QApplication.processEvents()
        self.resizeBars()

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def resizeEvent(self, event):
        """Resizes the outside changes dialogue if necessary"""
        QWidget.resizeEvent(self, event)
        self.resizeBars()

    def resizeBars(self):
        """Resize the bars if they are shown"""
        if not self.__outsideChangesBar.isHidden():
            self.__outsideChangesBar.resize()

    def __onReload(self):
        """Triggered when a request to reload the file is received"""
        self.reloadRequst.emit()

    def reloadAllNonModified(self):
        """Triggered when a request to reload all the
           non-modified files is received
        """
        self.reloadAllNonModifiedRequest.emit()

    def showOutsideChangesBar(self, allEnabled):
        """Shows the bar for the viewer for the user to choose the action"""
        self.setReloadDialogShown(True)
        self.__outsideChangesBar.showChoice(self.isModified(), allEnabled)

    def reload(self):
        """Called (from the editors manager) to reload the file"""
        # Re-read the file with updating the file timestamp
        self.loadFromFile(self.__fileName)

        # Hide the bar is necessery
        if not self.__outsideChangesBar.isHidden():
            self.__outsideChangesBar.hide()

        # Set the shown flag
        self.setReloadDialogShown(False)

    # Mandatory interface part is below

    def isModified(self):
        """Tells if the file is modified"""
        return False

    def getRWMode(self):
        """Tells if the file is read only"""
        return "RO"

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.PictureViewer

    def getLanguage(self):
        """Tells the content language"""
        return self.__viewer.formatInfo

    def getFileName(self):
        """Tells what file name of the widget content"""
        return self.__fileName

    def setFileName(self, path):
        """Sets the file name"""
        self.__fileName = path
        self.__shortName = os.path.basename(path)

    def getEncoding(self):
        """Tells the content encoding"""
        return self.__viewer.info

    def setEncoding(self, newEncoding):
        """Sets the encoding - not applicable for picture viewer"""

    def getShortName(self):
        """Tells the display name"""
        return self.__shortName

    def setShortName(self, name):
        """Sets the display name"""
        self.__shortName = name

    def isDiskFileModified(self):
        """Return True if the loaded file is modified"""
        if not os.path.isabs(self.__fileName):
            return False
        if not os.path.exists(self.__fileName):
            return True
        path = os.path.realpath(self.__fileName)
        return self.__diskModTime != os.path.getmtime(path) or \
            self.__diskSize != os.path.getsize(path)

    def doesFileExist(self):
        """Returns True if the loaded file still exists"""
        return os.path.exists(self.__fileName)

    def setReloadDialogShown(self, value=True):
        """Sets the new value of the flag which tells if the reloading
           dialogue has already been displayed
        """
        self.__reloadDlgShown = value

    def getReloadDialogShown(self):
        """Tells if the reload dialog has already been shown"""
        return self.__reloadDlgShown and \
            not self.__outsideChangesBar.isVisible()

    def updateModificationTime(self, fileName):
        """Updates the modification time"""
        path = os.path.realpath(fileName)
        self.__diskModTime = os.path.getmtime(path)
        self.__diskSize = os.path.getsize(path)

    def getVCSStatus(self):
        """Provides the VCS status"""
        return self.__vcsStatus

    def setVCSStatus(self, newStatus):
        """Sets the new VCS status"""
        self.__vcsStatus = newStatus



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2018  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""MD rendering widget"""


import logging
import os.path
from ui.qt import (Qt, QSize, QTimer, QToolBar, QWidget, QHBoxLayout,
                   QLabel, QVBoxLayout, QSizePolicy, QFrame,
                   pyqtSignal, QPrintDialog, QDialog, QAction, QPixmap,
                   QTextDocument)
from ui.texttabwidget import TextViewer
from utils.fileutils import isMarkdownMime
from utils.pixmapcache import getPixmap, getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from utils.diskvaluesrelay import getFilePosition
from utils.md import renderMarkdown


IDLE_TIMEOUT = 1500


class MDViewer(TextViewer):

    """Markdown rendered content viewer"""

    def __init__(self, parent):
        TextViewer.__init__(self, parent)

        self.setViewportMargins(10, 10, 10, 10)

        Settings().webResourceCache.sigResourceSaved.connect(
            self.onResourceSaved)
        Settings().plantUMLCache.sigRenderReady.connect(
            self.onPlantUMLRender)

    def getScrollbarPositions(self):
        """Provides the scrollbar positions"""
        hScrollBar = self.horizontalScrollBar()
        hsbValue = hScrollBar.value() if hScrollBar else 0

        vScrollBar = self.verticalScrollBar()
        vsbValue = vScrollBar.value() if vScrollBar else 0
        return hsbValue, vsbValue

    def setScrollbarPositions(self, hPos, vPos):
        """Sets the scrollbar positions for the view"""
        hsb = self.horizontalScrollBar()
        if hsb:
            hsb.setValue(hPos)
        vsb = self.verticalScrollBar()
        if vsb:
            vsb.setValue(vPos)

    def loadResource(self, resourceType, resourceURL):
        """Overloaded; by default the remote pixmaps are not loaded"""
        if resourceType == QTextDocument.ImageResource:
            url = resourceURL.toString()
            currentFileName = self._parentWidget.getFileName()
            if currentFileName:
                currentDir = os.path.dirname(currentFileName) + os.path.sep
                if url.startswith(currentDir):
                    url = url.replace(currentDir, '', 1)
            lowerUrl = url.lower()

            if url.startswith('plantuml:'):
                fName = url[9:]
                if os.path.exists(fName):
                    return QPixmap(fName)
                # I cannot return my picture. If I do that then the resource
                # will not be reloaded when the generated diagram is ready
                return None

            if url.startswith('cdm-image:'):
                try:
                    return getPixmap(url[10:])
                except Exception as exc:
                    logging.error('Unknown Codimension cache image: ' +
                                  url[10:])
                return None

            if lowerUrl.startswith('http:/') or lowerUrl.startswith('https:/'):
                if not lowerUrl.startswith('http://') and \
                   not lowerUrl.startswith('https://'):
                    url = url.replace(':/', '://', 1)
                fName = Settings().webResourceCache.getResource(
                    url, self._parentWidget.getUUID())
                if fName is not None:
                    try:
                        return QPixmap(fName)
                    except Exception as exc:
                        logging.error('Cannot use the image from ' + fName +
                                      ': ' + str(exc))
                return None
        return TextViewer.loadResource(self, resourceType, resourceURL)

    def onPlantUMLRender(self, uuid, fName):
        """Triggered when a diagram rendering is done"""
        self.onResourceSaved(None, uuid, fName)

    def onResourceSaved(self, url, uuid, fName):
        """Triggered when a pixmap is received asynchronously"""
        if uuid == self._parentWidget.getUUID():
            # Note: it is a hack! If the image is received after the document
            # is showed first time then its layout is wrong - not enough space
            # for the image and only a portion of the picture is shown.
            # The document re-layouting is required but there is no way to do
            # it directly. So there is this not-changing-anything call
            # which leads to re-layouting (update() does not help)
            # See here:
            # https://www.qtcentre.org/threads/6744-QTextEdit-and-delayed-image-loading
            self.setLineWrapColumnOrWidth(0)

    def terminate(self):
        """Called when a tab is to be closed"""
        Settings().webResourceCache.sigResourceSaved.disconnect(
            self.onResourceSaved)
        Settings().plantUMLCache.sigRenderReady.disconnect(
            self.onPlantUMLRender)


class MDTopBar(QFrame):

    """MD widget top bar at the top"""

    STATE_OK_UTD = 0        # Parsed OK, MD up to date
    STATE_OK_CHN = 1        # Parsed OK, MD changed
    STATE_BROKEN_UTD = 2    # Parsed with errors, MD up to date
    STATE_BROKEN_CHN = 3    # Parsed with errors, MD changed
    STATE_UNKNOWN = 4

    def __init__(self, parent):
        QFrame.__init__(self, parent)
        self.__infoIcon = None
        self.__warningsIcon = None
        self.__layout = None
        self.__createLayout()
        self.__currentIconState = self.STATE_UNKNOWN

    def __createLayout(self):
        """Creates the layout"""
        self.setFixedHeight(24)
        self.__layout = QHBoxLayout(self)
        self.__layout.setContentsMargins(0, 0, 0, 0)

        # Create info icon
        self.__infoIcon = QLabel(self)
        self.__infoIcon.setPixmap(getPixmap('cfunknown.png'))
        self.__layout.addWidget(self.__infoIcon)

        self.__warningsIcon = QLabel(self)
        self.__warningsIcon.setPixmap(getPixmap('cfwarning.png'))
        self.__layout.addWidget(self.__warningsIcon)

        self.clearWarnings()

        self.__spacer = QWidget(self)
        self.__spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.__spacer.setMinimumWidth(0)
        self.__layout.addWidget(self.__spacer)

    def clearWarnings(self):
        """Clears the warnings"""
        self.__warningsIcon.setVisible(False)
        self.__warningsIcon.setToolTip("")

    def setWarnings(self, warnings):
        """Sets the warnings"""
        self.__warningsIcon.setToolTip(
            'Markdown parser warnings:\n' + '\n'.join(warnings))
        self.__warningsIcon.setVisible(True)

    def clearErrors(self):
        """Clears all the errors"""
        self.__infoIcon.setToolTip('')

    def setErrors(self, errors):
        """Sets the errors"""
        self.__infoIcon.setToolTip(
            'Markdown parser errors:\n' + '\n'.join(errors))

    def updateInfoIcon(self, state):
        """Updates the information icon"""
        if state == self.__currentIconState:
            return

        if state == self.STATE_OK_UTD:
            self.__infoIcon.setPixmap(getPixmap('cfokutd.png'))
            self.__infoIcon.setToolTip("Markdown render is up to date")
            self.__currentIconState = self.STATE_OK_UTD
        elif state == self.STATE_OK_CHN:
            self.__infoIcon.setPixmap(getPixmap('cfokchn.png'))
            self.__infoIcon.setToolTip("Markdown render is not up to date; "
                                       "will be updated on idle")
            self.__currentIconState = self.STATE_OK_CHN
        elif state == self.STATE_BROKEN_UTD:
            self.__infoIcon.setPixmap(getPixmap('cfbrokenutd.png'))
            self.__infoIcon.setToolTip("Markdown render might be invalid "
                                       "due to invalid python code")
            self.__currentIconState = self.STATE_BROKEN_UTD
        elif state == self.STATE_BROKEN_CHN:
            self.__infoIcon.setPixmap(getPixmap('cfbrokenchn.png'))
            self.__infoIcon.setToolTip("Markdown render might be invalid; "
                                       "will be updated on idle")
            self.__currentIconState = self.STATE_BROKEN_CHN
        else:
            # STATE_UNKNOWN
            self.__infoIcon.setPixmap(getPixmap('cfunknown.png'))
            self.__infoIcon.setToolTip("Markdown render state is unknown")
            self.__currentIconState = self.STATE_UNKNOWN

    def getCurrentState(self):
        """Provides the current state"""
        return self.__currentIconState

    def resizeEvent(self, event):
        """Editor has resized"""
        QFrame.resizeEvent(self, event)


class MDWidget(QWidget):

    """The MD rendered content widget which goes along with the text editor"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, editor, parent):
        QWidget.__init__(self, parent)

        self.setVisible(False)

        self.__editor = editor
        self.__parentWidget = parent
        self.__connected = False

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)

        vLayout = QVBoxLayout()
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.setSpacing(0)

        # Make pylint happy
        self.__toolbar = None
        self.__topBar = None

        # Create the update timer
        self.__updateTimer = QTimer(self)
        self.__updateTimer.setSingleShot(True)
        self.__updateTimer.timeout.connect(self.process)

        vLayout.addWidget(self.__createTopBar())
        vLayout.addWidget(self.__createMDView())

        hLayout.addLayout(vLayout)
        hLayout.addWidget(self.__createToolbar())
        self.setLayout(hLayout)

        # Connect to the change file type signal
        self.__mainWindow = GlobalData().mainWindow
        editorsManager = self.__mainWindow.em
        editorsManager.sigFileTypeChanged.connect(self.__onFileTypeChanged)

    def __createToolbar(self):
        """Creates the toolbar"""
        self.__toolbar = QToolBar(self)
        self.__toolbar.setOrientation(Qt.Vertical)
        self.__toolbar.setMovable(False)
        self.__toolbar.setAllowedAreas(Qt.RightToolBarArea)
        self.__toolbar.setIconSize(QSize(16, 16))
        self.__toolbar.setFixedWidth(30)
        self.__toolbar.setContentsMargins(0, 0, 0, 0)

        # Some control buttons could be added later
        self.printButton = QAction(getIcon('printer.png'), 'Print', self)
        self.printButton.triggered.connect(self.__onPrint)
        self.__toolbar.addAction(self.printButton)

        return self.__toolbar

    def __createTopBar(self):
        """Creates the top bar"""
        self.__topBar = MDTopBar(self)
        return self.__topBar

    def __createMDView(self):
        """Creates the graphics view"""
        self.mdView = MDViewer(self)
        self.mdView.sigEscapePressed.connect(self.__onEsc)
        return self.mdView

    def __onPrint(self):
        """Print the markdown page"""
        dialog = QPrintDialog(self)
        if dialog.exec_() == QDialog.Accepted:
            printer = dialog.printer()
            self.mdView.print_(printer)

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def process(self):
        """Parses the content and displays the results"""
        if not self.__connected:
            self.__connectEditorSignals()

        renderedText, errors, warnings = renderMarkdown(self.getUUID(),
                                                        self.__editor.text,
                                                        self.getFileName())
        if errors:
            self.__topBar.updateInfoIcon(self.__topBar.STATE_BROKEN_UTD)
            self.__topBar.setErrors(errors)
            return
        if renderedText is None:
            self.__topBar.updateInfoIcon(self.__topBar.STATE_BROKEN_UTD)
            self.__topBar.setErrors(['Unknown markdown rendering error'])
            return

        # That will clear the error tooltip as well
        self.__topBar.updateInfoIcon(self.__topBar.STATE_OK_UTD)

        if warnings:
            self.__topBar.setWarnings(warnings)
        else:
            self.__topBar.clearWarnings()

        hsbValue, vsbValue = self.getScrollbarPositions()
        self.mdView.setHtml(renderedText)
        self.setScrollbarPositions(hsbValue, vsbValue)

    def __onFileTypeChanged(self, fileName, uuid, newFileType):
        """Triggered when a buffer content type has changed"""
        if self.getUUID() != uuid:
            return

        if not isMarkdownMime(newFileType):
            self.__disconnectEditorSignals()
            self.__updateTimer.stop()
            self.setVisible(False)
            self.__topBar.updateInfoIcon(self.__topBar.STATE_UNKNOWN)
            return

        # Update the bar and show it
        self.setVisible(True)
        self.process()

        # The buffer type change event comes when the content is loaded first
        # time. So this is a good point to restore the position
        _, _, _, hPos, vPos = getFilePosition(fileName)
        self.setScrollbarPositions(hPos, vPos)

    def terminate(self):
        """Called when a tab is to be closed"""
        self.mdView.terminate()
        self.mdView.deleteLater()

        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
        self.__updateTimer.deleteLater()

        self.__disconnectEditorSignals()

        editorsManager = self.__mainWindow.em
        editorsManager.sigFileTypeChanged.disconnect(self.__onFileTypeChanged)

        self.printButton.triggered.disconnect(self.__onPrint)
        self.printButton.deleteLater()

        self.__topBar.deleteLater()
        self.__toolbar.deleteLater()

        self.__editor = None
        self.__parentWidget = None

    def __connectEditorSignals(self):
        """When it is a python file - connect to the editor signals"""
        if not self.__connected:
            self.__editor.cursorPositionChanged.connect(
                self.__cursorPositionChanged)
            self.__editor.textChanged.connect(self.__onBufferChanged)
            self.__connected = True

    def __disconnectEditorSignals(self):
        """Disconnect the editor signals when the file is not a python one"""
        if self.__connected:
            self.__editor.cursorPositionChanged.disconnect(
                self.__cursorPositionChanged)
            self.__editor.textChanged.disconnect(self.__onBufferChanged)
            self.__connected = False

    def __cursorPositionChanged(self):
        """Cursor position changed"""
        # The timer should be reset only in case if the redrawing was delayed
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
            self.__updateTimer.start(IDLE_TIMEOUT)

    def __onBufferChanged(self):
        """Triggered to update status icon and to restart the timer"""
        self.__updateTimer.stop()
        if self.__topBar.getCurrentState() in [self.__topBar.STATE_OK_UTD,
                                               self.__topBar.STATE_OK_CHN,
                                               self.__topBar.STATE_UNKNOWN]:
            self.__topBar.updateInfoIcon(self.__topBar.STATE_OK_CHN)
        else:
            self.__topBar.updateInfoIcon(self.__topBar.STATE_BROKEN_CHN)
        self.__updateTimer.start(IDLE_TIMEOUT)

    def redrawNow(self):
        """Redraw the diagram regardless of the timer"""
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
        self.process()

    def getScrollbarPositions(self):
        """Provides the scrollbar positions"""
        return self.mdView.getScrollbarPositions()

    def setScrollbarPositions(self, hPos, vPos):
        """Sets the scrollbar positions for the view"""
        self.mdView.setScrollbarPositions(hPos, vPos)

    def getFileName(self):
        return self.__parentWidget.getFileName()

    def getUUID(self):
        return self.__parentWidget.getUUID()




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""VCS plugin support: plugin thread"""

import time
from collections import deque
from ui.qt import QThread, QMutex, QWaitCondition, pyqtSignal
from plugins.categories.vcsiface import VersionControlSystemInterface


# Indicator used by IDE to display errors while retrieving item status
IND_VCS_ERROR = -2


class VCSPluginThread(QThread):

    """Wrapper for the plugin thread"""

    VCSStatus = pyqtSignal(str, int, str)

    def __init__(self, plugin, parent=None):
        QThread.__init__(self, parent)

        self.__plugin = plugin
        self.__requestQueue = deque()

        self.__stopRequest = False
        self.__lock = QMutex()
        self.__condition = QWaitCondition()

    def run(self):
        """Thread loop"""
        while not self.__stopRequest:
            self.__lock.lock()
            while self.__requestQueue:
                path, flag = self.__requestQueue.pop()
                self.__lock.unlock()
                time.sleep(0.01)
                self.__processRequest(path, flag)
                if self.__stopRequest:
                    break
                self.__lock.lock()
            if self.__stopRequest:
                self.__lock.unlock()
                break
            self.__condition.wait(self.__lock)
            self.__lock.unlock()

    def __processRequest(self, path, flag):
        """Processes a single request. It must be exception safe."""
        try:
            statuses = self.__plugin.getObject().getStatus(path, flag)
            for status in statuses:
                if len(status) == 3:
                    self.VCSStatus.emit(path + status[0], status[1], status[2])
                else:
                    self.VCSStatus.emit(
                        path, IND_VCS_ERROR,
                        "The " + self.__plugin.getName() + " plugin "
                        "does not follow the getStatus() interface agreement")
        except Exception as exc:
            self.VCSStatus.emit(
                path, IND_VCS_ERROR,
                "Exception in " + self.__plugin.getName() +
                " plugin while retrieving VCS status: " + str(exc))
        except:
            self.VCSStatus.emit(
                path, IND_VCS_ERROR,
                "Unknown exception in " + self.__plugin.getName() +
                " plugin while retrieving VCS status")

    def addRequest(self, path, flag, urgent=False):
        """Adds a request to the queue"""
        self.__lock.lock()
        if urgent:
            self.__requestQueue.append((path, flag))
        else:
            self.__requestQueue.appendleft((path, flag))
        self.__lock.unlock()
        self.__condition.wakeAll()

    def stop(self):
        """Delivers the stop request"""
        self.__stopRequest = True
        self.__condition.wakeAll()

    def clearRequestQueue(self):
        """Clears the thread request queue"""
        self.__lock.lock()
        self.__requestQueue.clear()
        self.__lock.unlock()

    def canInitiateStatusRequestLoop(self):
        """Returns true if the is no jam in the request queue"""
        # It is a very rough test which seems however good enough.
        # First it is checked that there are no more than 10 (picked arbitrary)
        # not served yet requests in the queue and then it is checked if there
        # is at least one directory status request from a previous iteration.
        self.__lock.lock()
        if len(self.__requestQueue) > 10:
            self.__lock.unlock()
            return False
        for item in self.__requestQueue:
            if item[1] == VersionControlSystemInterface.REQUEST_DIRECTORY:
                self.__lock.unlock()
                return False
        self.__lock.unlock()
        return True



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Break points viewer"""

import logging
from ui.qt import (Qt, pyqtSignal, QSize, QSizePolicy, QFrame, QTreeView,
                   QHeaderView, QVBoxLayout, QSortFilterProxyModel,
                   QWidget, QAbstractItemView, QMenu, QAction, QModelIndex,
                   QCursor, QItemSelectionModel, QDialog, QToolBar)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.labels import HeaderFitLabel
from ui.spacers import ToolBarHSpacer, ToolBarExpandingSpacer
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from utils.project import CodimensionProject
from .editbreakpoint import BreakpointEditDialog
from .breakpoint import Breakpoint
from .bputils import getBreakpointLines
from .breakpointmodel import COLUMN_TEMPORARY, COLUMN_ENABLED, COLUMN_LOCATION


class BreakPointView(QTreeView):

    """Breakpoint viewer widget"""

    sigSelectionChanged = pyqtSignal(QModelIndex)

    def __init__(self, parent, bpointsModel):
        QTreeView.__init__(self, parent)

        self.__model = None
        self.setModel(bpointsModel)

        self.setItemsExpandable(False)
        self.setRootIsDecorated(False)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.__showContextMenu)
        self.doubleClicked.connect(self.__doubleClicked)

        self.__createPopupMenus()

    def setModel(self, model):
        """Sets the breakpoint model"""
        self.__model = model

        self.sortingModel = QSortFilterProxyModel()
        self.sortingModel.setSourceModel(self.__model)
        QTreeView.setModel(self, self.sortingModel)

        header = self.header()
        header.setSortIndicator(COLUMN_LOCATION, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)

        self.setSortingEnabled(True)
        self.layoutDisplay()

    def layoutDisplay(self):
        """Performs the layout operation"""
        self.__resizeColumns()
        self.__resort()

    def __resizeColumns(self):
        """Resizes the view when items get added, edited or deleted"""
        self.header().setStretchLastSection(True)
        self.header().resizeSections(QHeaderView.ResizeToContents)
        self.header().resizeSection(COLUMN_TEMPORARY, 22)
        self.header().resizeSection(COLUMN_ENABLED, 22)

    def __resort(self):
        """Resorts the tree"""
        self.model().sort(self.header().sortIndicatorSection(),
                          self.header().sortIndicatorOrder())

    def toSourceIndex(self, index):
        """Converts an index to a source index"""
        return self.sortingModel.mapToSource(index)

    def __fromSourceIndex(self, sindex):
        """Convert a source index to an index"""
        return self.sortingModel.mapFromSource(sindex)

    def __setRowSelected(self, index, selected=True):
        """Selects a row"""
        if not index.isValid():
            return

        if selected:
            flags = QItemSelectionModel.SelectionFlags(
                QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
        else:
            flags = QItemSelectionModel.SelectionFlags(
                QItemSelectionModel.Deselect | QItemSelectionModel.Rows)
        self.selectionModel().select(index, flags)

    def __createPopupMenus(self):
        """Generate the popup menu"""
        self.menu = QMenu()
        self.__editAct = self.menu.addAction(
            getIcon('bpprops.png'), "Edit...", self.__editBreak)
        self.__jumpToCodeAct = self.menu.addAction(
            getIcon('gotoline.png'), "Jump to code", self.__showSource)
        self.menu.addSeparator()
        self.__enableAct = self.menu.addAction(
            getIcon('bpenable.png'), "Enable", self.enableBreak)
        self.__enableAllAct = self.menu.addAction(
            getIcon('bpenableall.png'), "Enable all", self.enableAllBreaks)
        self.menu.addSeparator()
        self.__disableAct = self.menu.addAction(
            getIcon('bpdisable.png'), "Disable", self.disableBreak)
        self.__disableAllAct = self.menu.addAction(
            getIcon('bpdisableall.png'), "Disable all", self.disableAllBreaks)
        self.menu.addSeparator()
        self.__delAct = self.menu.addAction(
            getIcon('bpdel.png'), "Delete", self.deleteBreak)
        self.__delAllAct = self.menu.addAction(
            getIcon('bpdelall.png'), "Delete all", self.deleteAllBreaks)

    def __showContextMenu(self, _):
        """Shows the context menu"""
        index = self.currentIndex()
        if not index.isValid():
            return
        sindex = self.toSourceIndex(index)
        if not sindex.isValid():
            return
        bpoint = self.__model.getBreakPointByIndex(sindex)
        if not bpoint:
            return

        enableCount, disableCount = self.__model.getCounts()

        self.__editAct.setEnabled(True)
        self.__enableAct.setEnabled(not bpoint.isEnabled())
        self.__disableAct.setEnabled(bpoint.isEnabled())
        self.__jumpToCodeAct.setEnabled(True)
        self.__delAct.setEnabled(True)
        self.__enableAllAct.setEnabled(disableCount > 0)
        self.__disableAllAct.setEnabled(enableCount > 0)
        self.__delAllAct.setEnabled(enableCount + disableCount > 0)

        self.menu.popup(QCursor.pos())

    def __doubleClicked(self, index):
        """Handles the double clicked signal"""
        if not index.isValid():
            return

        sindex = self.toSourceIndex(index)
        if not sindex.isValid():
            return

        # Jump to the code
        bpoint = self.__model.getBreakPointByIndex(sindex)
        fileName = bpoint.getAbsoluteFileName()
        line = bpoint.getLineNumber()
        self.jumpToCode(fileName, line)

    @staticmethod
    def jumpToCode(fileName, line):
        """Jumps to the source code"""
        editorsManager = GlobalData().mainWindow.editorsManager()
        editorsManager.openFile(fileName, line)
        editor = editorsManager.currentWidget().getEditor()
        editor.gotoLine(line)
        editorsManager.currentWidget().setFocus()

    def __editBreak(self):
        """Handle the edit breakpoint context menu entry"""
        index = self.currentIndex()
        if index.isValid():
            self.__editBreakpoint(index)

    def __editBreakpoint(self, index):
        """Edits a breakpoint"""
        sindex = self.toSourceIndex(index)
        if sindex.isValid():
            bpoint = self.__model.getBreakPointByIndex(sindex)
            if not bpoint:
                return

            dlg = BreakpointEditDialog(bpoint)
            if dlg.exec_() == QDialog.Accepted:
                newBpoint = dlg.getData()
                if newBpoint == bpoint:
                    return
                self.__model.setBreakPointByIndex(sindex, newBpoint)
                self.layoutDisplay()

    def __setBpEnabled(self, index, enabled):
        """Sets the enabled status of a breakpoint"""
        sindex = self.toSourceIndex(index)
        if sindex.isValid():
            self.__model.setBreakPointEnabledByIndex(sindex, enabled)

    def enableBreak(self):
        """Handles the enable breakpoint context menu entry"""
        index = self.currentIndex()
        self.__setBpEnabled(index, True)
        self.__resizeColumns()
        self.__resort()

    def enableAllBreaks(self):
        """Handles the enable all breakpoints context menu entry"""
        index = self.model().index(0, 0)
        while index.isValid():
            self.__setBpEnabled(index, True)
            index = self.indexBelow(index)
        self.__resizeColumns()
        self.__resort()

    def disableBreak(self):
        """Handles the disable breakpoint context menu entry"""
        index = self.currentIndex()
        self.__setBpEnabled(index, False)
        self.__resizeColumns()
        self.__resort()

    def disableAllBreaks(self):
        """Handles the disable all breakpoints context menu entry"""
        index = self.model().index(0, 0)
        while index.isValid():
            self.__setBpEnabled(index, False)
            index = self.indexBelow(index)
        self.__resizeColumns()
        self.__resort()

    def deleteBreak(self):
        """Handles the delete breakpoint context menu entry"""
        index = self.currentIndex()
        sindex = self.toSourceIndex(index)
        if sindex.isValid():
            self.__model.deleteBreakPointByIndex(sindex)

    def deleteAllBreaks(self):
        """Handles the delete all breakpoints context menu entry"""
        self.__model.deleteAll()

    def __showSource(self):
        """Handles the goto context menu entry"""
        index = self.currentIndex()
        self.__doubleClicked(index)

    def highlightBreakpoint(self, fname, lineno):
        """Handles the clientLine signal"""
        sindex = self.__model.getBreakPointIndex(fname, lineno)
        if sindex.isValid():
            return

        index = self.__fromSourceIndex(sindex)
        if index.isValid():
            self.__clearSelection()
            self.__setRowSelected(index, True)

    def __getSelectedItemsCount(self):
        """Provides the count of items selected"""
        count = len(self.selectedIndexes()) / (self.__model.columnCount() - 1)
        # column count is 1 greater than selectable
        return count

    def selectionChanged(self, selected, deselected):
        """The slot is called when the selection has changed"""
        if selected.indexes():
            self.sigSelectionChanged.emit(selected.indexes()[0])
        else:
            self.sigSelectionChanged.emit(QModelIndex())
        QTreeView.selectionChanged(self, selected, deselected)


class BreakPointViewer(QWidget):

    """Implements the break point viewer for a debugger"""

    def __init__(self, parent, bpointsModel):
        QWidget.__init__(self, parent)

        self.__currentItem = None
        self.__createLayout(bpointsModel)

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)
        GlobalData().project.sigProjectAboutToUnload.connect(
            self.__onProjectAboutToUnload)
        self.bpointsList.sigSelectionChanged.connect(self.__onSelectionChanged)
        bpointsModel.sigBreakpoinsChanged.connect(self.__onModelChanged)

    def setFocus(self):
        """Sets the widget focus"""
        self.bpointsList.setFocus()

    def __createLayout(self, bpointsModel):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__breakpointLabel = HeaderFitLabel(self)
        self.__breakpointLabel.setText('Breakpoints')
        self.__breakpointLabel.setSizePolicy(QSizePolicy.Expanding,
                                             QSizePolicy.Fixed)
        self.__breakpointLabel.setMinimumWidth(10)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(18, 18))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__breakpointLabel)

        self.bpointsList = BreakPointView(self, bpointsModel)

        self.__editButton = QAction(
            getIcon('bpprops.png'), "Edit breakpoint properties", self)
        self.__editButton.triggered.connect(self.__onEdit)
        self.__editButton.setEnabled(False)

        self.__jumpToCodeButton = QAction(
            getIcon('gotoline.png'), "Jump to the code", self)
        self.__jumpToCodeButton.triggered.connect(self.__onJumpToCode)
        self.__jumpToCodeButton.setEnabled(False)

        self.__enableButton = QAction(
            getIcon('bpenable.png'), "Enable selected breakpoint", self)
        self.__enableButton.triggered.connect(self.__onEnableDisable)
        self.__enableButton.setEnabled(False)

        self.__disableButton = QAction(
            getIcon('bpdisable.png'), "Disable selected breakpoint", self)
        self.__disableButton.triggered.connect(self.__onEnableDisable)
        self.__disableButton.setEnabled(False)

        self.__enableAllButton = QAction(
            getIcon('bpenableall.png'), "Enable all the breakpoint", self)
        self.__enableAllButton.triggered.connect(self.__onEnableAll)
        self.__enableAllButton.setEnabled(False)

        self.__disableAllButton = QAction(
            getIcon('bpdisableall.png'), "Disable all the breakpoint", self)
        self.__disableAllButton.triggered.connect(self.__onDisableAll)
        self.__disableAllButton.setEnabled(False)

        self.__delButton = QAction(
            getIcon('delitem.png'), "Delete selected breakpoint", self)
        self.__delButton.triggered.connect(self.__onDel)
        self.__delButton.setEnabled(False)

        self.__delAllButton = QAction(
            getIcon('bpdelall.png'), "Delete all the breakpoint", self)
        self.__delAllButton.triggered.connect(self.__onDelAll)
        self.__delAllButton.setEnabled(False)


        # Toolbar
        self.toolbar = QToolBar()
        self.toolbar.setOrientation(Qt.Horizontal)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedHeight(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.__editButton)
        self.toolbar.addAction(self.__jumpToCodeButton)
        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 5))
        self.toolbar.addAction(self.__enableButton)
        self.toolbar.addAction(self.__enableAllButton)
        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 5))
        self.toolbar.addAction(self.__disableButton)
        self.toolbar.addAction(self.__disableAllButton)
        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 5))
        self.toolbar.addWidget(ToolBarExpandingSpacer(self.toolbar))
        self.toolbar.addAction(self.__delButton)
        self.toolbar.addWidget(ToolBarHSpacer(self.toolbar, 5))
        self.toolbar.addAction(self.__delAllButton)

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.toolbar)
        verticalLayout.addWidget(self.bpointsList)

    def clear(self):
        """Clears the content"""
        self.__onDelAll()
        self.__updateBreakpointsLabel()
        self.__currentItem = None

    def __updateBreakpointsLabel(self):
        """Updates the breakpoints header label"""
        enableCount, \
        disableCount = self.bpointsList.model().sourceModel().getCounts()
        total = enableCount + disableCount
        if total > 0:
            self.__breakpointLabel.setText("Breakpoints (total: " +
                                           str(total) + ")")
        else:
            self.__breakpointLabel.setText("Breakpoints")

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what != CodimensionProject.CompleteProject:
            return

        self.clear()
        model = self.bpointsList.model().sourceModel()
        project = GlobalData().project
        if project.isLoaded():
            bpoints = project.breakpoints
        else:
            bpoints = Settings().breakpoints

        for bpoint in bpoints:
            newBpoint = Breakpoint()
            try:
                if not newBpoint.deserialize(bpoint):
                    # Non valid
                    continue
            except:
                continue
            # Need to check if it still points to a breakable line
            line = newBpoint.getLineNumber()
            fileName = newBpoint.getAbsoluteFileName()
            breakableLines = getBreakpointLines(fileName, None, True)
            if breakableLines is not None and line in breakableLines:
                model.addBreakpoint(newBpoint)
            else:
                logging.warning("Breakpoint at " + fileName + ":" +
                                str(line) + " does not point to a breakable "
                                "line anymore (the file is invalid or was "
                                "modified outside of the "
                                "IDE etc.). The breakpoint is deleted.")

    def __onProjectAboutToUnload(self):
        """Triggered before the project is unloaded"""
        self.__serializeBreakpoints()

    def __serializeBreakpoints(self):
        """Saves the breakpoints into a file"""
        model = self.bpointsList.model().sourceModel()

        project = GlobalData().project
        if project.isLoaded():
            project.breakpoints = model.serialize()
        else:
            Settings().breakpoints = model.serialize()

    def __onSelectionChanged(self, index):
        """Triggered when the current item is changed"""
        if index.isValid():
            srcModel = self.bpointsList.model().sourceModel()
            sindex = self.bpointsList.toSourceIndex(index)
            self.__currentItem = srcModel.getBreakPointByIndex(sindex)
        else:
            self.__currentItem = None
        self.__updateButtons()

    def __updateButtons(self):
        """Updates the buttons status"""
        enableCount, \
        disableCount = self.bpointsList.model().sourceModel().getCounts()

        if self.__currentItem is None:
            self.__editButton.setEnabled(False)
            self.__enableButton.setEnabled(False)
            self.__disableButton.setEnabled(False)
            self.__jumpToCodeButton.setEnabled(False)
            self.__delButton.setEnabled(False)
        else:
            self.__editButton.setEnabled(True)
            self.__enableButton.setEnabled(not self.__currentItem.isEnabled())
            self.__disableButton.setEnabled(self.__currentItem.isEnabled())
            self.__jumpToCodeButton.setEnabled(True)
            self.__delButton.setEnabled(True)

        self.__enableAllButton.setEnabled(disableCount > 0)
        self.__disableAllButton.setEnabled(enableCount > 0)
        self.__delAllButton.setEnabled(enableCount + disableCount > 0)

    def __onEnableDisable(self):
        """Triggered when a breakpoint should be enabled/disabled"""
        if self.__currentItem is not None:
            if self.__currentItem.isEnabled():
                self.bpointsList.disableBreak()
            else:
                self.bpointsList.enableBreak()

    def __onEdit(self):
        """Triggered when a breakpoint should be edited"""
        if self.__currentItem is None:
            return

        dlg = BreakpointEditDialog(self.__currentItem)
        if dlg.exec_() == QDialog.Accepted:
            newBpoint = dlg.getData()
            if newBpoint == self.__currentItem:
                return
            model = self.bpointsList.model().sourceModel()
            index = model.getBreakPointIndex(
                self.__currentItem.getAbsoluteFileName(),
                self.__currentItem.getLineNumber())
            model.setBreakPointByIndex(index, newBpoint)
            self.bpointsList.layoutDisplay()

    def __onJumpToCode(self):
        """Triggered when should jump to source"""
        if self.__currentItem is None:
            return
        self.bpointsList.jumpToCode(self.__currentItem.getAbsoluteFileName(),
                                    self.__currentItem.getLineNumber())

    def __onEnableAll(self):
        """Triggered when all the breakpoints should be enabled"""
        self.bpointsList.enableAllBreaks()

    def __onDisableAll(self):
        """Triggered when all the breakpoints should be disabled"""
        self.bpointsList.disableAllBreaks()

    def __onDel(self):
        """Triggered when a breakpoint should be deleted"""
        if self.__currentItem is not None:
            self.bpointsList.deleteBreak()

    def __onDelAll(self):
        """Triggered when all the breakpoints should be deleted"""
        self.bpointsList.deleteAllBreaks()

    def __onModelChanged(self):
        """Triggered when something has changed in any of the breakpoints"""
        self.__updateBreakpointsLabel()
        self.__updateButtons()
        self.bpointsList.layoutDisplay()

        self.__serializeBreakpoints()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""Pyflakes margin"""

import math
import logging
from html import escape
import qutepart
from qutepart.margins import MarginBase
from radon.complexity import cc_rank
from ui.qt import QWidget, QPainter, QToolTip
from utils.misc import extendInstance
from utils.globals import GlobalData
from utils.fileutils import isPythonMime
from utils.pixmapcache import getPixmap

# Pixmaps for the margin - populated once
MARKS = {}

class CDMFlakesMargin(QWidget):

    """Pyflakes area widget"""

    EXC_MARK = 1
    CURRENT_MARK = 2
    FLAKES_MARK = 3

    COMPLEXITY_A_MARK = ord('A')
    COMPLEXITY_B_MARK = ord('B')
    COMPLEXITY_C_MARK = ord('C')
    COMPLEXITY_D_MARK = ord('D')
    COMPLEXITY_E_MARK = ord('E')
    COMPLEXITY_F_MARK = ord('F')

    def __init__(self, parent):
        QWidget.__init__(self, parent)

        extendInstance(self, MarginBase)
        MarginBase.__init__(self, parent, "cdm_flakes_margin", 1)
        self.setMouseTracking(True)

        self.__messages = {}
        self.__ccMessages = {}
        self.__bgColor = GlobalData().skin['flakesMarginPaper']
        self.__noTooltip = True

        self.currentDebugLine = None
        self.excptionLine = None

        if not MARKS:
            MARKS[self.CURRENT_MARK] = getPixmap('dbgcurrentmarker.png')
            MARKS[self.EXC_MARK] = getPixmap('dbgexcptmarker.png')
            MARKS[self.FLAKES_MARK] = getPixmap('pyflakesmsgmarker.png')
            MARKS[self.COMPLEXITY_A_MARK] = getPixmap('complexity-a.png')
            MARKS[self.COMPLEXITY_B_MARK] = getPixmap('complexity-b.png')
            MARKS[self.COMPLEXITY_C_MARK] = getPixmap('complexity-c.png')
            MARKS[self.COMPLEXITY_D_MARK] = getPixmap('complexity-d.png')
            MARKS[self.COMPLEXITY_E_MARK] = getPixmap('complexity-e.png')
            MARKS[self.COMPLEXITY_F_MARK] = getPixmap('complexity-f.png')

            for item in MARKS:
                if MARKS[item].height() != MARKS[item].width():
                    logging.error('analysis margin pixmap needs to be square')

        self.myUUID = None
        if hasattr(self._qpart._parent, 'getUUID'):
            self.myUUID = self._qpart._parent.getUUID()

        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.connect(self.__onFileTypeChanged)
        self._qpart.blockCountChanged.connect(self.update)

    def paintEvent(self, event):
        """Paints the margin"""
        painter = QPainter(self)
        painter.fillRect(event.rect(), self.__bgColor)
        oneLineHeight = self._qpart.fontMetrics().height()

        block = self._qpart.firstVisibleBlock()
        geometry = self._qpart.blockBoundingGeometry(block)
        blockBoundingGeometry = geometry.translated(
            self._qpart.contentOffset())
        top = blockBoundingGeometry.top()
        # bottom = top + blockBoundingGeometry.height()

        for block in qutepart.iterateBlocksFrom(block):
            height = self._qpart.blockBoundingGeometry(block).height()
            if top > event.rect().bottom():
                break
            if block.isVisible():
                lineNo = block.blockNumber() + 1
                pixmap = None
                if lineNo == self.excptionLine:
                    pixmap = MARKS[self.EXC_MARK]
                elif lineNo == self.currentDebugLine:
                    pixmap = MARKS[self.CURRENT_MARK]
                elif self.isBlockMarked(block):
                    if lineNo in self.__ccMessages:
                        pixmap = MARKS[self.__ccMessages[lineNo][1]]
                    else:
                        pixmap = MARKS[self.FLAKES_MARK]

                if pixmap:
                    xPos = 0
                    yPos = top
                    pixmapSide = self.width()   # Pixmap is square!
                    if oneLineHeight >= pixmapSide:
                        # More than enough vertical space, width is fixed
                        yPos += math.ceil((oneLineHeight - pixmapSide) / 2)
                    else:
                        # Not enough vertical space, width is fixed
                        xPos += math.ceil((pixmapSide - oneLineHeight) / 2)
                        pixmapSide = oneLineHeight
                    painter.drawPixmap(xPos, yPos,
                                       pixmapSide, pixmapSide, pixmap)
            top += height

    def mouseMoveEvent(self, event):
        """Tooltips for the marks"""
        if not self.__noTooltip:
            blockNumber = self._qpart.cursorForPosition(
                event.pos()).blockNumber()
            lineno = blockNumber + 1
            msg = None

            if lineno == self.excptionLine:
                msg = 'Exception line'
            elif lineno == self.currentDebugLine:
                msg = 'Current debugger line'
            elif lineno in self.__ccMessages:
                msg = self.__ccMessages[lineno][0]
            elif lineno in self.__messages:
                msg = ''
                for part in self.__messages[lineno]:
                    if msg:
                        msg += '<br/>'
                    msg += escape(part)
                msg = "<p style='white-space:pre'>" + msg + "</p>"

            if msg:
                QToolTip.showText(event.globalPos(), msg)
            else:
                QToolTip.hideText()
        return QWidget.mouseMoveEvent(self, event)

    @staticmethod
    def width():
        """Desired width"""
        return 16

    def setBackgroundColor(self, color):
        """Sets the new background color"""
        if self.__bgColor != color:
            self.__bgColor = color
            self.update()

    def __onFileTypeChanged(self, _, uuid, newFileType):
        """Triggered on the changed file type"""
        if uuid == self.myUUID:
            MarginBase.setVisible(self, isPythonMime(newFileType))

    def __onBlockCountChanged(self):
        """Triggered when the block count changed"""
        self.__noTooltip = True
        self.update()

    def clearAnalysisMessages(self):
        """Clears all the messages"""
        self.__messages = {}
        self.__ccMessages = {}
        self.clear()

    def clearDebugMarks(self):
        """Clears all debug marks"""
        self.excptionLine = None
        self.currentDebugLine = None
        self.clear()
        self.update()

    def setCurrentDebugLine(self, currentDebugLine):
        """Sets the current debug line"""
        self.currentDebugLine = currentDebugLine
        self.excptionLine = None
        self.update()

    def setExceptionLine(self, exceptionLine):
        """Sets the exception line"""
        self.excptionLine = exceptionLine
        self.currentDebugLine = None
        self.update()

    def setAnalysisMessages(self, messages, ccMessages):
        """Sets a new set of messages"""
        self.__messages = dict(messages)

        for lineno in self.__messages:
            if lineno > 0:
                self.setBlockValue(
                    self._qpart.document().findBlockByNumber(lineno - 1), 1)

        self.__ccMessages = {}
        for item in ccMessages:
            if item.lineno not in self.__messages:
                complexity = cc_rank(item.complexity)
                if complexity != 'A':
                    msg = 'Cyclomatic complexity is ' + complexity + \
                          ' (value: ' + str(item.complexity) + ')'
                    self.__ccMessages[item.lineno] = (msg, ord(complexity))
                    self.setBlockValue(
                        self._qpart.document().findBlockByNumber(
                            item.lineno - 1), 1)

        self.__noTooltip = False
        self.update()

    def onClose(self):
        """The editor is going to be closed"""
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigFileTypeChanged.disconnect(self.__onFileTypeChanged)
        self._qpart.blockCountChanged.disconnect(self.update)

        # Release the editor reference
        self._qpart = None



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

""" Provides versions of various components used by codimension """

from os.path import abspath
from distutils.spawn import find_executable
from subprocess import getstatusoutput
import logging
import pkg_resources
from .plantumlcache import JAR_PATH


def getPackageVersionAndLocation(name):
    """Provides a package version"""
    try:
        return pkg_resources.get_distribution(name).version, \
               pkg_resources.get_distribution(name).location
    except pkg_resources.DistributionNotFound as exc:
        logging.error(str(exc))
        return None, None


def getCodimensionVersion():
    """Provides the IDE version"""
    from .globals import GlobalData
    import sys
    return GlobalData().version, abspath(sys.argv[0])


def getPythonInterpreterVersion():
    """Provides the python interpreter version"""
    import sys
    return ".".join([str(sys.version_info.major),
                     str(sys.version_info.minor),
                     str(sys.version_info.micro)]), sys.executable


def getQtVersion():
    """Provides the Qt version"""
    from ui.qt import QT_VERSION_STR
    return QT_VERSION_STR


def getGraphvizVersion():
    """Provides the graphviz version"""
    from .globals import GlobalData
    if not GlobalData().graphvizAvailable:
        return "Not installed", None

    path = find_executable("dot")
    if not path:
        return "Not installed", None

    try:
        status, output = getstatusoutput(path + ' -V')
        if status != 0:
            return "Not installed", None

        for line in output.splitlines():
            # E.g. dot - graphviz version 2.26.3 (20100126.1600)
            line = line.strip()
            if line.startswith("dot - graphviz version "):
                line = line.replace("dot - graphviz version ", "")
                parts = line.split(" ")
                if len(parts) == 2 and parts[0][0].isdigit():
                    return parts[0], path
    except:
        return "Not installed", None
    return "could not determine", path


def getJavaVersion():
    """Provides java version"""
    from .globals import GlobalData
    if not GlobalData().javaAvailable:
        return "Not installed", None

    path = find_executable("java")
    if not path:
        return "Not installed", None

    try:
        status, output = getstatusoutput(path + ' -version')
        if status != 0:
            return "Not installed", None

        for line in output.splitlines():
            # E.g. openjdk version "1.8.0_212"
            line = line.strip()
            if 'version' in line:
                ver = line.split('version', 1)[1].strip()
                if ver.startswith('"') and ver.endswith('"'):
                    ver = ver[1:-1]
                return ver, path
    except:
        return "Not installed", None
    return "could not determine", path


def getPlantUMLVersion():
    """Provides the plantUML version"""
    if JAR_PATH is None:
        return "n/a", None

    try:
        status, output = getstatusoutput('java -jar ' + JAR_PATH + ' -version')
        if status != 0:
            return "n/a", None

        for line in output.splitlines():
            # PlantUML version 1.2019.05 (Sat Apr 20 11:45:36 GMT-05:00 2019)
            line = line.strip()
            if line.startswith('PlantUML version'):
                return line.split('version', 1)[1].strip(), JAR_PATH
    except:
        return "n/a", None
    return "could not determine", JAR_PATH


def getComponentInfo():
    """Provides major codimension components information"""
    components = []
    # Each item contains: <pretty name>, <version>,
    #                     <url>, <patched>, <license name>,
    #                     <license url>
    # A list is used to have some kind of priority ordering
    version, path = getCodimensionVersion()
    components.append(("Codimension IDE", version,
                       "http://codimension.org/", None,
                       "GPL-3.0",
                       "http://www.gnu.org/licenses/gpl-3.0.html",
                       path))
    version, path = getPackageVersionAndLocation('cdmpyparser')
    components.append(("Codimension python parser", version,
                       "http://codimension.org/", None,
                       "GPL-3.0",
                       "http://www.gnu.org/licenses/gpl-3.0.html",
                       path))
    version, path = getPackageVersionAndLocation('cdmcfparser')
    components.append(("Codimension python control flow parser",
                       version,
                       "http://codimension.org/", None,
                       "GPL-3.0",
                       "http://www.gnu.org/licenses/gpl-3.0.html",
                       path))
    version, path = getPythonInterpreterVersion()
    components.append(("Python interpreter", version,
                       "http://www.python.org/", None,
                       "Open source license",
                       "http://www.python.org/psf/license/",
                       path))
    version, path = getPackageVersionAndLocation('PyQt5')
    components.append(("PyQt", version,
                       "http://www.riverbankcomputing.com/software/pyqt/intro",
                       None, "GPL-2.0/GPL-3.0/Commercial/Embedded",
                       "http://www.riverbankcomputing.com/"
                       "software/pyqt/license",
                       path))
    version, path = getPackageVersionAndLocation('qutepart')
    components.append(("qutepart", version,
                       "https://github.com/andreikop/qutepart",
                       None,
                       "LGPL-2.1",
                       "http://www.gnu.org/licenses/lgpl-2.1.html",
                       path))
    components.append(("Qt", getQtVersion(),
                       "http://qt-project.org/", None,
                       "LGPL-2.1/Commercial",
                       "http://www.gnu.org/licenses/lgpl-2.1.html",
                       None))
    version, path = getPackageVersionAndLocation('pyflakes')
    components.append(("pyflakes", version,
                       "https://pypi.python.org/pypi/pyflakes", None,
                       "pyflakes license", "see the package",
                       path))
    version, path = getPackageVersionAndLocation('python-magic')
    components.append(("python-magic", version,
                       "https://pypi.python.org/pypi/python-magic/", None,
                       "MIT license",
                       "https://opensource.org/licenses/MIT",
                       path))
    version, path = getGraphvizVersion()
    components.append(("graphviz", version,
                       "http://www.graphviz.org/", None,
                       "Eclipse Public License 1.0",
                       "http://www.graphviz.org/License.php",
                       path))
    version, path = getPackageVersionAndLocation('gprof2dot')
    components.append(("gprof2dot", version,
                       "https://github.com/jrfonseca/gprof2dot", None,
                       "LGPL", "http://www.gnu.org/licenses/lgpl.html",
                       path))
    version, path = getPackageVersionAndLocation('yapsy')
    components.append(("yapsy", version,
                       "http://yapsy.sourceforge.net", None,
                       "BSD License",
                       "http://opensource.org/licenses/bsd-license.php",
                       path))
    version, path = getPackageVersionAndLocation('jedi')
    components.append(("jedi", version,
                       "http://jedi.readthedocs.io", None,
                       "MIT License",
                       "https://opensource.org/licenses/MIT",
                       path))
    version, path = getPackageVersionAndLocation('radon')
    components.append(("radon", version,
                       "https://radon.readthedocs.org/", None,
                       "MIT License",
                       "https://opensource.org/licenses/MIT",
                       path))
    version, path = getPackageVersionAndLocation('vulture')
    components.append(("vulture", version,
                       "https://github.com/jendrikseipp/vulture", None,
                       "MIT License",
                       "https://opensource.org/licenses/MIT",
                       path))
    version, path = getPackageVersionAndLocation('mistune')
    components.append(("mistune", version,
                       "https://github.com/lepture/mistune", None,
                       "BSD 3-Clause License",
                       "https://opensource.org/licenses/BSD-3-Clause",
                       path))
    version, path = getJavaVersion()
    components.append(("java", version, None, None, None, None, path))
    version, path = getPlantUMLVersion()
    components.append(('PlantUML', version,
                       'http://plantuml.com/', None,
                       'MIT License',
                       'https://opensource.org/licenses/MIT',
                       path))
    return components



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Control flow UI widget: handling mouse events"""

# pylint: disable=W0702

from sys import maxsize
from flowui.scopeitems import ScopeCellElement
from flowui.cellelement import CellElement
from flowui.auxitems import RubberBandItem
from ui.qt import Qt, QTransform, QRect, QSize, QGraphicsScene


# The default mouse behavior of the QT library is sometimes inconsistent.
# For example, selecting of the items is done on mousePressEvent however
# adding items is done on mouseReleaseEvent.
# One more problem is that the QT library does not support hierarchical
# relationships between the items. I.e. if an outer item is selected then
# it makes no sense to let nested items addable.
# The last problem is that there are proxy items (like badges) which are not
# supported out of the box obviously.
# So the whole mouse[Press,Release]Event members are overridden.
# The selection always works on mouse release event. Mouse press will do
# nothing.


RUBBER_BAND_MIN_SIZE = 5

class CFSceneMouseMixin:

    """Encapsulates mouse clicks handling and related functionality"""

    def __init__(self):
        self.lmbOrigin = None
        self.rubberBand = None

    @staticmethod
    def __getLogicalItem(item):
        """Provides a logical item or None"""
        if item is None:
            return None
        if item.isProxyItem():
            # This is for an SVG, a badge, a text and a connector
            return item.getProxiedItem()
        if not item.scopedItem():
            return item
        # Here: it is a scope item. Need to map/suppress in some cases
        if item.subKind in [ScopeCellElement.DECLARATION]:
            # Need to map to the top left item because the out
            return item.getTopLeftItem()
        if item.subKind in [ScopeCellElement.COMMENT,
                            ScopeCellElement.DOCSTRING]:
            # No mapping
            return item
        # ScopeCellElement.TOP_LEFT is to be ignored
        return None

    def __getCanvas(self):
        """Provides a reference to the canvas (used for rubber band)"""
        for item in self.items():
            if item.kind == CellElement.FILE_SCOPE:
                return item.canvas
        return None

    def __createRubberBand(self, event):
        """Creates the rubber band rectangle and shows it"""
        # Canvas == None means that there is no module item on the canvas
        # which is a case when there is exactly nothing drawn. The only
        # possiblity is that a broken file was open so there is nothing to
        # select anyway
        canvas = self.__getCanvas()
        if canvas is not None:
            self.lmbOrigin = event.scenePos().toPoint()
            self.rubberBand = RubberBandItem(canvas)
            self.addItem(self.rubberBand)
            self.rubberBand.setGeometry(QRect(self.lmbOrigin, QSize()))
            self.rubberBand.hide()

    def __destroyRubberBand(self):
        """Destroys the rubber band selection rectangle"""
        if self.rubberBand is not None:
            try:
                # Sometimes there is a race of the rubber band destruction
                self.rubberBand.hide()
            except:
                pass
            self.rubberBand = None
        self.lmbOrigin = None

    def mousePressEvent(self, event):
        """Handles mouse press event"""
        button = event.button()
        if button not in [Qt.LeftButton, Qt.RightButton]:
            self.clearSelection()
            event.accept()
            return

        if button == Qt.RightButton:
            item = self.itemAt(event.scenePos(), QTransform())
            logicalItem = self.__getLogicalItem(item)
            if logicalItem is None:
                # Not a selectable item or out of the items
                self.clearSelection()
            elif not logicalItem.isSelected():
                self.clearSelection()
                logicalItem.setSelected(True)

            # Bring up a context menu
            self.onContextMenu(event)
        else:
            # Here: this is LMB
            self.__createRubberBand(event)

        event.accept()
        return

    def mouseMoveEvent(self, event):
        """Handles mouse movement"""
        if self.lmbOrigin and self.rubberBand:
            # Draw the rubber band selection rectangle
            rect = QRect(self.lmbOrigin, event.scenePos().toPoint())
            self.rubberBand.setGeometry(rect.normalized())
            if not self.__isRubberBandVisible():
                if abs(rect.left() - rect.right()) >= RUBBER_BAND_MIN_SIZE or \
                   abs(rect.top() - rect.bottom()) >= RUBBER_BAND_MIN_SIZE:
                    self.rubberBand.show()
        QGraphicsScene.mouseMoveEvent(self, event)

    def __isRubberBandVisible(self):
        """Tells if the rubber band is on the screen"""
        try:
            # Sometimes there is a race somewhere so that the wrapped C++
            # object is already destroyed, so make the test in a try block
            return self.rubberBand and self.rubberBand.isVisible()
        except:
            self.__destroyRubberBand()
            return False

    def mouseReleaseEvent(self, event):
        """Handles the mouse release event"""
        button = event.button()

        if button not in [Qt.LeftButton, Qt.RightButton]:
            event.accept()
            return

        if button == Qt.RightButton:
            event.accept()
            return

        # Here: left mouse button
        if self.__isRubberBandVisible():
            # Detect intersections
            self.clearSelection()
            for item in self.items():
                if not isinstance(item, CellElement):
                    continue
                if item.isProxyItem():
                    continue
                if item.scopedItem():
                    if item.subKind not in [ScopeCellElement.DECLARATION,
                                            ScopeCellElement.COMMENT,
                                            ScopeCellElement.DOCSTRING]:
                        continue

                    if item.subKind == ScopeCellElement.DECLARATION:
                        item = item.getTopLeftItem()

                # The call must be done on item. (not on rubberBand.)
                if item.collidesWithItem(self.rubberBand,
                                         Qt.ContainsItemBoundingRect):
                    self.addToSelection(item)
        else:
            item = self.itemAt(event.scenePos(), QTransform())
            logicalItem = self.__getLogicalItem(item)

            if logicalItem is None:
                self.clearSelection()
            else:
                modifiers = event.modifiers()
                if modifiers == Qt.NoModifier:
                    if logicalItem.kind in [CellElement.INDEPENDENT_DOC,
                                            CellElement.LEADING_DOC,
                                            CellElement.ABOVE_DOC] and \
                       item.isProxyItem():
                        if logicalItem.cmlRef.link is None:
                            self.clearSelection()
                            logicalItem.setSelected(True)
                        else:
                            logicalItem.mouseClickLinkIcon()
                    else:
                        self.clearSelection()
                        logicalItem.setSelected(True)
                elif modifiers == Qt.ControlModifier:
                    if logicalItem.isSelected():
                        logicalItem.setSelected(False)
                    else:
                        # Item is not selected and should be added or ignored
                        self.addToSelection(logicalItem)
                # The alt modifier works for the whole app window on
                # Ubuntu, so it cannot really be used...
                elif modifiers == Qt.ShiftModifier:
                    self.clearSelection()

                    # Here: add comments
                    if self.isOpenGroupItem(item):
                        self.addToSelection(item)
                    else:
                        for itemForRef in self.findItemsForRef(logicalItem.ref):
                            self.addToSelection(itemForRef)

        self.__destroyRubberBand()
        event.accept()

    def addToSelection(self, item):
        """Adds an item to the current selection"""
        if self.isNestedInSelected(item):
            # Ignore the selection request
            return

        self.deselectNested(item)
        item.setSelected(True)

    def isNestedInSelected(self, itemToSelect):
        """Tells if the item is already included into some other selected"""
        toSelectBegin, toSelectEnd = self.__getItemVisualBeginEnd(itemToSelect)

        items = self.selectedItems()
        for item in items:
            isGroup = self.isOpenGroupItem(item)
            if not item.scopedItem() and not isGroup:
                continue
            if item.scopedItem():
                if item.subKind != ScopeCellElement.TOP_LEFT:
                    continue

            itemBegin, itemEnd = self.__getItemVisualBeginEnd(item)
            if toSelectBegin > itemBegin and toSelectEnd < itemEnd:
                return True
        return False

    def deselectNested(self, itemToSelect):
        """Deselects all the nested items if needed"""
        isGroup = itemToSelect.kind == CellElement.OPENED_GROUP_BEGIN
        if not itemToSelect.scopedItem() and not isGroup:
            # The only scope items and groups require
            # deselection of the nested items
            return
        if itemToSelect.scopedItem():
            if itemToSelect.subKind != ScopeCellElement.TOP_LEFT:
                # Scope docstrings and side comments cannot include anything
                return

        toSelectBegin, toSelectEnd = self.__getItemVisualBeginEnd(itemToSelect)
        items = self.selectedItems()
        for item in items:
            itemBegin, itemEnd = self.__getItemVisualBeginEnd(item)
            if itemBegin >= toSelectBegin and itemEnd <= toSelectEnd:
                item.setSelected(False)

    def __getItemVisualBeginEnd(self, item):
        """Provides the item visual begin and end"""
        if item.scopedItem():
            if item.subKind == ScopeCellElement.COMMENT:
                return item.ref.sideComment.begin, item.ref.sideComment.end
            if item.subKind == ScopeCellElement.DOCSTRING:
                return item.ref.docstring.begin, item.ref.docstring.end
            # File scope differs from the other scopes
            if item.kind == CellElement.FILE_SCOPE:
                if item.ref.docstring:
                    return item.ref.docstring.begin, item.ref.body.end
                return item.ref.body.begin, item.ref.body.end

            # try, while, for are special
            if item.kind in [CellElement.TRY_SCOPE,
                             CellElement.FOR_SCOPE,
                             CellElement.WHILE_SCOPE]:
                lastSuiteItem = item.ref.suite[-1]
                return item.ref.body.begin, lastSuiteItem.end

            return item.ref.body.begin, item.ref.end

        # Here: not a scope item.
        if item.kind == CellElement.ASSERT:
            if item.ref.message is not None:
                end = item.ref.message.end
            elif item.ref.test is not None:
                end = item.ref.test.end
            else:
                end = item.ref.body.end
            return item.ref.body.begin, end
        if item.kind == CellElement.RAISE:
            if item.ref.value is not None:
                end = item.ref.value.end
            else:
                end = item.ref.body.end
            return item.ref.body.begin, end
        if item.kind == CellElement.RETURN:
            if item.ref.value is not None:
                end = item.ref.value.end
            else:
                end = item.ref.body.end
            return item.ref.body.begin, end

        # All the rest
        begin, end = item.getAbsPosRange()
        return begin, end

    def findItemsForRef(self, ref):
        """Provides graphics items for the given ref"""
        result = []
        for item in self.items():
            if not isinstance(item, CellElement):
                continue
            if item.scopedItem():
                if item.subKind not in [ScopeCellElement.TOP_LEFT,
                                        ScopeCellElement.COMMENT,
                                        ScopeCellElement.DOCSTRING]:
                    continue
            if not self.isOpenGroupItem(item):
                if hasattr(item, "ref"):
                    if item.ref is ref:
                        result.append(item)
        return result

    @staticmethod
    def isOpenGroupItem(item):
        """True if it is an open group item"""
        if hasattr(item, 'kind'):
            return item.kind == CellElement.OPENED_GROUP_BEGIN
        return False

    def getNearestItem(self, absPos, line, pos):
        """Provides a logical item and the distance.

        The item is the closest to the specified absPos, line:pos.
        The distance to the item (0 - within the item).
        line and pos are 1-based.
        """
        del pos     # unused argument
        candidates = []
        distance = maxsize

        for item in self.items():
            if not isinstance(item, CellElement):
                continue
            if item.isProxyItem():
                continue

            dist = item.getLineDistance(line)
            if dist == maxsize:
                continue    # Not really an option
            if dist < distance:
                distance = dist
                candidates = [item]
            elif dist == distance:
                candidates.append(item)

        count = len(candidates)
        if count == 0:
            return None, maxsize
        if count == 1:
            return self.__getLogicalItem(candidates[0]), distance

        # Here: more than one item with an equal distance
        #       There are two cases here: 0 and non zero distance
        if distance != 0:
            # It is pretty much not important which one to pick.
            # Let it be the first foun item
            return self.__getLogicalItem(candidates[0]), distance

        # This is a zero line distance, so a candidate should be the one with
        # the shortest position distance
        candidate = None
        distance = maxsize
        for item in candidates:
            dist = item.getDistance(absPos)
            if dist == 0:
                return self.__getLogicalItem(item), 0
            if dist < distance:
                distance = dist
                candidate = item
        return self.__getLogicalItem(candidate), distance



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CML utilities"""

from sys import maxsize
from cdmcfparser import (IF_FRAGMENT, FOR_FRAGMENT, WHILE_FRAGMENT,
                         TRY_FRAGMENT, CONTROL_FLOW_FRAGMENT, CLASS_FRAGMENT,
                         FUNCTION_FRAGMENT, CML_COMMENT_FRAGMENT)
from utils.colorfont import buildColor, cssLikeColor


def escapeCMLTextValue(src):
    """Escapes the string before inserting it to the code (CML value)"""
    dst = ''
    for char in src:
        if char == '\n':
            dst += '\\n'
        elif char in ('"', '\\'):
            dst += '\\' + char
        else:
            dst += char
    return dst


def unescapeCMLTextValue(src):
    """Removes escaping from the value received from the code"""
    dst = ''
    lastIndex = len(src) - 1
    index = 0
    while index <= lastIndex:
        if src[index] == '\\' and index < lastIndex:
            if src[index + 1] == '\\':
                dst += '\\'
                index += 1
            elif src[index + 1] == 'n':
                dst += '\n'
                index += 1
            elif src[index + 1] == '"':
                dst += '"'
                index += 1
            else:
                dst += '\\'     # forgiving the other escape characters
        else:
            dst += src[index]
        index += 1
    return dst


def appendTextItem(name, value):
    """Provides a line to append"""
    if value:
        if ' ' in value or '"' in value:
            return ' ' + name + '="' + escapeCMLTextValue(value) + '"'
        return ' ' + name + '=' + escapeCMLTextValue(value)
    return ''


class CMLColorBase:

    """Covers the common colors for CML comments"""

    def __init__(self):
        self.bgColor = None         # background color
        self.fgColor = None         # foreground color
        self.border = None          # border color

    def populateColors(self):
        """Populate colors"""
        if "bg" in self.ref.properties:
            self.bgColor = buildColor(self.ref.properties["bg"])
        if "fg" in self.ref.properties:
            self.fgColor = buildColor(self.ref.properties["fg"])
        if "border" in self.ref.properties:
            self.border = buildColor(self.ref.properties["border"])

    @staticmethod
    def generateColors(background, foreground, border):
        """Generates the colors part"""
        parts = []
        if background is not None:
            parts.append('bg=' + cssLikeColor(background))
        if foreground is not None:
            parts.append('fg=' + cssLikeColor(foreground))
        if border is not None:
            parts.append('border=' + cssLikeColor(border))
        return ' '.join(parts)



class CMLCommentBase:

    """Base class for all the CML comments"""

    def __init__(self, ref=None):
        self.ref = ref
        self.kind = CML_COMMENT_FRAGMENT

    def validateRecordType(self, code):
        """Validates the record type"""
        if self.ref.recordType != code:
            raise Exception("Invalid CML comment type. "
                            "Expected: '" + code + "'. Received: '" +
                            self.ref.recordType + "'.")

    def __isSideComment(self, editor):
        """True if it is a side comment"""
        # The only first part needs to be checked
        firstPart = self.ref.parts[0]
        # Editor has 0-based lines
        leftStripped = editor.lines[firstPart.beginLine - 1].lstrip()
        return not leftStripped.startswith(firstPart.getContent())

    def removeFromText(self, editor):
        """Removes the comment from the code"""
        # Note: it is supposed that the 'with editor:' is done outside.
        #       This is because the required changes could be done for more
        #       than one place.
        if editor is None:
            return

        isSideComment = self.__isSideComment(editor)

        oldLine, oldPos = editor.cursorPosition
        line = self.ref.endLine
        while line >= self.ref.beginLine:
            if isSideComment:
                raise Exception("Side CML comments removal "
                                "has not been implemented yet")
            else:
                # Editor has 0-based lines
                del editor.lines[line - 1]
                if oldLine >= line - 1:
                    oldLine -= 1
            line -= 1

        editor.cursorPosition = oldLine, oldPos

    def getLineRange(self):
        """Provides the line range"""
        # The CML comments may appear in a suite and range may be asked from
        # them to decide if the grouping is available
        return [self.ref.parts[0].beginLine,
                self.ref.parts[-1].endLine]

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return [self.ref.parts[0].begin, self.ref.parts[-1].end]


class CMLsw(CMLCommentBase):

    """Covers the 'if' statement CML SW (switch branches) comments"""

    CODE = "sw"

    def __init__(self, ref):
        CMLCommentBase.__init__(self, ref)
        self.validate()

    def validate(self):
        """Validates the CML comment"""
        self.validateRecordType(CMLsw.CODE)
        CMLVersion.validate(self.ref)

    @staticmethod
    def description():
        """Provides the CML comment description"""
        return "The '" + CMLsw.CODE + \
               "' comment is used for 'if' and 'elif' statements " \
               "to switch default branch location i.e. to have " \
               "the 'No' branch at the right.\n" \
               "Supported properties: none\n\n" \
               "Example:\n" \
               "# cml 1 " + CMLsw.CODE

    @staticmethod
    def generate(pos=1):
        """Generates a complete line to be inserted"""
        return " " * (pos - 1) + "# cml 1 sw"


class CMLcc(CMLCommentBase, CMLColorBase):

    """Covers 'Custom Colors' spec for most of the items"""

    CODE = "cc"

    def __init__(self, ref):
        CMLCommentBase.__init__(self, ref)
        CMLColorBase.__init__(self)
        self.validate()

    def validate(self):
        """Validates the CML comment"""
        self.validateRecordType(CMLcc.CODE)
        CMLVersion.validate(self.ref)

        self.populateColors()
        if self.bgColor is None and \
                self.fgColor is None and self.border is None:
            raise Exception("The '" + CMLcc.CODE + "' CML comment does not "
                            "supply neither background nor foreground color "
                            "nor border color")

    @staticmethod
    def description():
        """Provides the CML comment description"""
        return "The '" + CMLcc.CODE + \
               "' comment is used for custom colors of most of " \
               "the graphics items.\n" \
               "Supported properties:\n" \
               "- color properties as described in the common section\n\n" \
               "Example:\n" \
               "# cml 1 " + CMLcc.CODE + \
               " bg=#f6f4e4 fg=#000 border=#fff"

    @staticmethod
    def generate(background, foreground, border, pos=1):
        """Generates a complete line to be inserted"""
        return  ' ' * (pos - 1) + '# cml 1 cc ' + \
                CMLColorBase.generateColors(background, foreground, border)


class CMLdoc(CMLCommentBase, CMLColorBase):

    """Covers the 'documentation link' comment"""

    CODE = "doc"

    def __init__(self, ref):
        CMLCommentBase.__init__(self, ref)
        CMLColorBase.__init__(self)
        self.link = None    # file path, abs or relative
        self.anchor = None  # id of this comment to let a link from the doc
        self.title = None   # text to display
        self.validate()

    def validate(self):
        """Validates the CMD doc comment"""
        self.validateRecordType(CMLdoc.CODE)
        CMLVersion.validate(self.ref)

        self.populateColors()

        self.link = self.ref.properties.get('link', None)
        self.anchor = self.ref.properties.get('anchor', None)
        self.title = self.ref.properties.get('title', '')

        if self.link is None and self.anchor is None:
            raise Exception("The '" + CMLdoc.CODE +
                            "' CML comment supplies neither "
                            "a link nor an anchor")
        if self.anchor:
            if ' ' in self.anchor or '\t' in self.anchor:
                raise Exception("The '" + CMLdoc.CODE +
                                "' CML comment anchor attribute "
                                "must contain neither spaces nor tabs")

    @staticmethod
    def description():
        """Provides the CML comment description"""
        return "The '" + CMLdoc.CODE + \
            "' comment is used to provide a link to a documentation.\n" \
            "Supported properties:\n" \
            "- 'link': link to the appropriate documentation\n" \
            "- 'anchor': this ID could be used to provide a link to this comment\n" \
            "- 'title': what to display on graphics\n" \
            "- color properties as described in the common section\n\n" \
            "Example:\n" \
            "# cml 1 " + CMLdoc.CODE + " link=file:doc/mydoc.md title=\"See more\""

    @staticmethod
    def generate(link, anchor, title, background, foreground, border, pos=1):
        """Generates a complete line to be inserted"""
        res = ' ' * (pos - 1) + '# cml 1 ' + CMLdoc.CODE + \
              appendTextItem('link', link) + \
              appendTextItem('anchor', anchor) + \
              appendTextItem('title', title)
        colorPart = CMLColorBase.generateColors(background, foreground, border)
        if colorPart:
            return res + ' ' + colorPart
        return res

    def getTitle(self):
        """Provides unescaped text"""
        return unescapeCMLTextValue(self.title)

    def updateCustomColors(self, editor, bgcolor, fgcolor, bordercolor):
        """Updates the custom colors"""
        firstCommentLine = self.ref.parts[0].beginLine
        pos = self.ref.parts[0].beginPos
        newCommentLine = CMLdoc.generate(self.link, self.anchor, self.title,
                                         bgcolor, fgcolor, bordercolor, pos)
        self.removeFromText(editor)
        editor.insertLines(newCommentLine, firstCommentLine)

    def removeCustomColors(self, editor):
        """Removes the custom colors"""
        self.updateCustomColors(editor, None, None, None)



class CMLrt(CMLCommentBase):

    """Covers 'Replace text' comment"""

    CODE = "rt"

    def __init__(self, ref):
        CMLCommentBase.__init__(self, ref)
        self.text = None
        self.validate()

    def validate(self):
        """Validates the CML rt comment"""
        self.validateRecordType(CMLrt.CODE)
        CMLVersion.validate(self.ref)

        self.text = self.ref.properties.get("text", None)

        if self.text is None:
            raise Exception("The '" + CMLrt.CODE +
                            "' CML comment does not supply text")

    @staticmethod
    def description():
        """Provides the CML comment description"""
        return "The '" + CMLrt.CODE + \
               "' comment is used for replacing the text of most of " \
               "the graphics items.\n" \
               "Supported properties:\n" \
               "- 'text': text to be shown instead of the real code\n\n" \
               "Example:\n" \
               "# cml 1 " + CMLrt.CODE + " text=\"Reset the dictionary\""

    @staticmethod
    def generate(txt, pos=1):
        """Generates a complete line to be inserted"""
        return " " * (pos - 1) + "# cml 1 " + CMLrt.CODE + \
               appendTextItem('text', txt)

    def getText(self):
        """Provides unescaped text"""
        if self.text is None:
            return None
        return unescapeCMLTextValue(self.text)


class CMLgb(CMLCommentBase, CMLColorBase):

    """Covers the 'group begin' comment"""

    CODE = 'gb'

    def __init__(self, ref):
        CMLCommentBase.__init__(self, ref)
        CMLColorBase.__init__(self)
        self.id = None
        self.title = None
        self.validate()

    def validate(self):
        """Validates the CML gb comment"""
        self.validateRecordType(CMLgb.CODE)
        CMLVersion.validate(self.ref)

        self.populateColors()

        self.title = self.ref.properties.get('title', None)
        if 'id' in self.ref.properties:
            self.id = self.ref.properties['id'].strip()

        if not self.id:
            raise Exception("The '" + CMLgb.CODE +
                            "' CML comment does not supply id")

    @staticmethod
    def description():
        """Provides the CML comment description"""
        return "The '" + CMLgb.CODE + \
               "' is used to indicate the beginning of the visual group. It " \
               "needs a counterpart " + CMLge.CODE + " CML comment which " \
               "indicates the end of the visual group.\n" \
               "Supported properties:\n" \
               "- 'title': title to be shown when the group is collapsed\n" \
               "- 'id': unique identifier of the visual group\n" \
               "- color properties as described in the common section\n\n" \
               "Example:\n" \
               "# cml 1 " + CMLgb.CODE + " id=\"1234-5678-444444\" " \
               "title=\"MD5 calculation\""

    @staticmethod
    def generate(groupid, title, background, foreground, border, pos=1):
        """Generates a complete line to be inserted"""
        res = ' ' * (pos - 1) + '# cml 1 ' + CMLgb.CODE + \
              appendTextItem('id', groupid) + \
              appendTextItem('title', title)
        colorPart = CMLColorBase.generateColors(background, foreground, border)
        if colorPart:
            return res + ' ' + colorPart
        return res

    def getTitle(self):
        """Provides unescaped title"""
        if self.title is None:
            return None
        return unescapeCMLTextValue(self.title)

    def updateTitle(self, editor, newTitle):
        """Updates the group title in the editor"""
        if newTitle == self.title:
            return

        firstCommentLine = self.ref.parts[0].beginLine
        lastCommentLine = self.ref.parts[-1].endLine
        if self.title is None:
            # Need to add the title= attribute to the last comment line
            lastLine = editor.lines[lastCommentLine - 1]
            newLastLine = lastLine.rstrip() + ' title="' + \
                escapeCMLTextValue(newTitle) + '"'
            editor.lines[lastCommentLine - 1] = newLastLine
            return

        # The title was there, so re-generate the comment line, remove the old
        # one and insert a new one
        pos = self.ref.parts[0].beginPos
        newCommentLine = CMLgb.generate(self.id, newTitle, self.bgColor,
                                        self.fgColor, self.border, pos)
        self.removeFromText(editor)
        editor.insertLines(newCommentLine, firstCommentLine)

    def updateCustomColors(self, editor, bgcolor, fgcolor, bordercolor):
        """Updates the custom colors"""
        firstCommentLine = self.ref.parts[0].beginLine
        pos = self.ref.parts[0].beginPos
        newCommentLine = CMLgb.generate(self.id, self.title, bgcolor,
                                        fgcolor, bordercolor, pos)
        self.removeFromText(editor)
        editor.insertLines(newCommentLine, firstCommentLine)

    def removeCustomColors(self, editor):
        """Removes the custom colors"""
        self.updateCustomColors(editor, None, None, None)



class CMLge(CMLCommentBase):

    """Covers the 'group end' comment"""

    CODE = 'ge'

    def __init__(self, ref):
        CMLCommentBase.__init__(self, ref)
        self.id = None
        self.validate()

    def validate(self):
        """Validates the CML ge comment"""
        self.validateRecordType(CMLge.CODE)
        CMLVersion.validate(self.ref)

        if 'id' in self.ref.properties:
            self.id = self.ref.properties['id'].strip()

        if not self.id:
            raise Exception("The '" + CMLge.CODE +
                            "' CML comment does not supply id")

    @staticmethod
    def description():
        """Provides the CML comment description"""
        return "The '" + CMLge.CODE + \
               "' is used to indicate the end of the visual group. It " \
               "needs a counterpart " + CMLgb.CODE + " CML comment which " \
               "indicates the beginning of the visual group.\n" \
               "Supported properties:\n" \
               "- 'id': unique identifier of the visual group\n\n" \
               "Example:\n" \
               "# cml 1 " + CMLge.CODE + " id=\"1234-5678-444444\""

    @staticmethod
    def generate(groupid, pos=1):
        """Generates a complete line to be inserted"""
        res = ' ' * (pos - 1) + '# cml 1 ' + CMLge.CODE
        if ' ' in groupid:
            res += ' id="' + groupid + '"'
        else:
            res += ' id=' + groupid
        return res



class CMLVersion:

    """Describes the current CML version"""

    VERSION = 1     # Current CML version
    COMMENT_TYPES = {CMLsw.CODE: CMLsw,
                     CMLcc.CODE: CMLcc,
                     CMLrt.CODE: CMLrt,
                     CMLgb.CODE: CMLgb,
                     CMLge.CODE: CMLge,
                     CMLdoc.CODE: CMLdoc}

    def __init__(self):
        pass

    @staticmethod
    def validate(cmlComment):
        """Valides the vestion"""
        if cmlComment.version > CMLVersion.VERSION:
            raise Exception("The CML comment version " +
                            str(cmlComment.version) +
                            " is not supported. Max supported version is " +
                            str(CMLVersion.VERSION))

    @staticmethod
    def find(cmlComments, cmlType):
        """Finds the CML comment"""
        for comment in cmlComments:
            if hasattr(comment, "CODE"):
                if comment.CODE == cmlType.CODE:
                    return comment
        return None

    @staticmethod
    def getType(cmlComment):
        """Provides the CML comment type"""
        try:
            return CMLVersion.COMMENT_TYPES[cmlComment.recordType]
        except KeyError:
            return None

    @staticmethod
    def validateCMLComments(item, validGroups, allGroupId,
                            pickLeadingComments=True):
        """Validates recursively all the CML items in the control flow.

        Replaces the recognized CML comments from the module with their higher
        level counterparts.
        Returns back a list of warnings. Also populates a list of valid groups.
        """
        scopeGroupStack = []    # [(id, lineBegin, lineEnd), ...]

        warnings = []
        if pickLeadingComments:
            # This is a special case for CML comments for the module.
            # They need to be considered as belonging to the module suite.
            if hasattr(item, "leadingCMLComments"):
                warnings += CMLVersion.validateCMLList(
                    item.leadingCMLComments, True, scopeGroupStack,
                    validGroups, allGroupId)

        # Some items are containers
        if item.kind == IF_FRAGMENT:
            for index, part in enumerate(item.parts):
                if index == 0:
                    # The very first part
                    warnings += CMLVersion.validateCMLList(
                        part.leadingCMLComments, True, scopeGroupStack,
                        validGroups, allGroupId)
                else:
                    warnings += CMLVersion.validateCMLList(
                        part.leadingCMLComments, False, None, None, None,
                        'elif or else parts')
                warnings += CMLVersion.validateCMLComments(
                    part, validGroups, allGroupId, False)
        elif item.kind in [FOR_FRAGMENT, WHILE_FRAGMENT]:
            if item.elsePart:
                warnings += CMLVersion.validateCMLList(
                    item.elsePart.leadingCMLComments, False, None, None, None,
                    'loop else parts')
                warnings += CMLVersion.validateCMLComments(
                    item.elsePart, validGroups, allGroupId, False)
        elif item.kind == TRY_FRAGMENT:
            if item.elsePart:
                warnings += CMLVersion.validateCMLList(
                    item.elsePart.leadingCMLComments, False,
                    None, None, None, 'try else parts')
                warnings += CMLVersion.validateCMLComments(
                    item.elsePart, validGroups, allGroupId, False)
            if item.finallyPart:
                warnings += CMLVersion.validateCMLList(
                    item.finallyPart.leadingCMLComments, False,
                    None, None, None, 'try finally parts')
                warnings += CMLVersion.validateCMLComments(
                    item.finallyPart, validGroups, allGroupId, False)
            for part in item.exceptParts:
                warnings += CMLVersion.validateCMLList(
                    part.leadingCMLComments, False,
                    None, None, None, 'try except parts')
                warnings += CMLVersion.validateCMLComments(
                    part, validGroups, allGroupId, False)
        elif item.kind in [CONTROL_FLOW_FRAGMENT,
                           CLASS_FRAGMENT, FUNCTION_FRAGMENT]:
            if item.docstring:
                warnings += CMLVersion.validateCMLList(
                    item.docstring.leadingCMLComments, False,
                    None, None, None, 'docstrings')
                warnings += CMLVersion.validateCMLList(
                    item.docstring.sideCMLComments, False,
                    None, None, None, 'docstrings')

            if item.kind in [CLASS_FRAGMENT, FUNCTION_FRAGMENT]:
                if item.decorators:
                    for decorator in item.decorators:
                        warnings += CMLVersion.validateCMLList(
                            decorator.leadingCMLComments, False,
                            None, None, None, 'decorators')
                        warnings += CMLVersion.validateCMLList(
                            decorator.sideCMLComments, False,
                            None, None, None, 'decorators')

        if hasattr(item, "sideCMLComments"):
            warnings += CMLVersion.validateCMLList(
                item.sideCMLComments, False, None, None, None, 'side comments')

        if hasattr(item, "suite"):
            for index, nestedItem in enumerate(item.suite):
                if nestedItem.kind == CML_COMMENT_FRAGMENT:
                    # independent CML comment
                    warn, replace = CMLVersion.validateCMLComment(nestedItem)
                    if replace is not None:
                        item.suite[index] = replace
                        if replace.CODE in [CMLgb.CODE, CMLge.CODE]:
                            CMLVersion.handleScopeGroup(replace,
                                                        scopeGroupStack,
                                                        warnings,
                                                        validGroups,
                                                        allGroupId)
                    if warn is not None:
                        warnings.append(warn)
                    continue

                if hasattr(nestedItem, "leadingCMLComments"):
                    warnings += CMLVersion.validateCMLList(
                        nestedItem.leadingCMLComments, True, scopeGroupStack,
                        validGroups, allGroupId)
                if nestedItem.kind == IF_FRAGMENT:
                    warnings += CMLVersion.validateCMLList(
                        nestedItem.parts[0].leadingCMLComments, True,
                        scopeGroupStack, validGroups, allGroupId)

                warnings += CMLVersion.validateCMLComments(nestedItem,
                                                           validGroups,
                                                           allGroupId, False)

        for group in scopeGroupStack:
            warnings.append((group[1], -1, 'CML ' + CMLgb.CODE + ' comment '
                             'does not have a matching CML ' + CMLge.CODE +
                             ' comment'))

        return warnings

    @staticmethod
    def handleScopeGroup(cmlComment, groupStack, warnings,
                         validGroups, allGroupId):
        """Processes the cml grouping comments"""
        line = cmlComment.ref.parts[0].beginLine
        if cmlComment.CODE == CMLgb.CODE:
            groupStack.append((cmlComment.id, line))
            if cmlComment.id:
                allGroupId.add(cmlComment.id)
            return
        if cmlComment.CODE == CMLge.CODE:
            if not groupStack:
                warnings.append((line, -1,
                                 'CML ' + CMLge.CODE + ' comment '
                                 'without CML ' + CMLgb.CODE + ' comment'))
                return
            if groupStack[-1][0] != cmlComment.id:
                warnings.append((line, -1,
                                 'CML ' + CMLge.CODE + ' comment id does not '
                                 'match the previous CML ' + CMLgb.CODE +
                                 ' comment id at line ' +
                                 str(groupStack[-1][1])))
                return
            validGroups.append((cmlComment.id, groupStack[-1][1], line))
            groupStack.pop()

    @staticmethod
    def validateCMLList(comments, pickGroups, groupStack,
                        validGroups, allGroupId, itemName=None):
        """Validates the CML comments in the provided list (internal use)"""
        warnings = []
        if comments:
            for index, item in enumerate(comments):
                if isinstance(item, CMLCommentBase):
                    continue

                warn, replace = CMLVersion.validateCMLComment(item)
                if replace is not None:
                    comments[index] = replace
                    if replace.CODE in [CMLgb.CODE, CMLge.CODE]:
                        if pickGroups:
                            CMLVersion.handleScopeGroup(replace,
                                                        groupStack,
                                                        warnings,
                                                        validGroups,
                                                        allGroupId)
                        else:
                            line = replace.ref.parts[0].beginLine
                            pos = replace.ref.parts[0].beginPos
                            warn = (line, pos, 'Groups are not allowed for ' +
                                    itemName)
                if warn is not None:
                    warnings.append(warn)
        return warnings

    @staticmethod
    def validateCMLComment(cmlComment):
        """Validates a CML comment (internal use)"""
        warning = None
        highLevel = None
        cmlType = CMLVersion.getType(cmlComment)
        if cmlType:
            try:
                highLevel = cmlType(cmlComment)
            except Exception as exc:
                line = cmlComment.parts[0].beginLine
                pos = cmlComment.parts[0].beginPos
                warning = (line, pos, "Invalid CML comment: " + str(exc))
        else:
            line = cmlComment.parts[0].beginLine
            pos = cmlComment.parts[0].beginPos
            warning = (line, pos,
                       "CML comment type '" + cmlComment.recordType +
                       "' is not supported")
        return warning, highLevel

    @staticmethod
    def getFirstLine(comments):
        """Provides the first line of the comment"""
        # The list may contain raw comments and high level comments
        line = maxsize
        if comments:
            if hasattr(comments[0], "ref"):
                # High level CML comment
                return comments[0].ref.parts[0].beginLine
            # Raw CML comment
            if comments[0].parts:
                return comments[0].parts[0].beginLine
        return line

    @staticmethod
    def description():
        """Provides the common parameters for various CML comments"""
        return "Color properties supported by the '" + CMLcc.CODE + \
               "' and the '" + CMLgb.CODE + "' comments:\n" \
               "- 'bg': background color for the item\n" \
               "- 'fg': foreground color for the item\n" \
               "- 'border': border color for the item\n" \
               "Color spec formats:\n" \
               "- '#hhh': hexadecimal RGB\n" \
               "- '#hhhh': hexadecimal RGBA\n" \
               "- '#hhhhhh': hexadecimal RRGGBB\n" \
               "- '#hhhhhhhh': hexadecimal RRGGBBAA\n" \
               "- 'ddd,ddd,ddd': decimal RGB\n" \
               "- 'ddd,ddd,ddd,ddd': decimal RGBA"



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Helper redirector class"""

from .qt import QObject, pyqtSignal


class Redirector(QObject):

    """Helper class used to redirect stdout and stderr to the log window"""

    appendToStdout = pyqtSignal(str)
    appendToStderr = pyqtSignal(str)

    def __init__(self, isStdout):
        QObject.__init__(self)
        self.__isStdout = isStdout

    def flush(self):
        """Flushes the buffered data - to conform the interface"""

    def write(self, message):
        """Writes the given data"""
        message = message.rstrip('\n')
        if self.__isStdout:
            self.appendToStdout.emit(message)
        else:
            self.appendToStderr.emit(message)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Various kinds of completers"""


from .qt import QDir, QStringListModel, QCompleter, QDirModel


class FileCompleter(QCompleter):

    """Completer for file names"""

    def __init__(self, parent=None,
                 completionMode=QCompleter.PopupCompletion,
                 showHidden=False):
        QCompleter.__init__(self, parent)
        self.__model = QDirModel(self)

        if showHidden:
            filters = QDir.Filters(QDir.Dirs | QDir.Files | QDir.Drives |
                                   QDir.AllDirs | QDir.Hidden)
        else:
            filters = QDir.Filters(QDir.Dirs | QDir.Files |
                                   QDir.Drives | QDir.AllDirs)
        self.__model.setFilter(filters)

        self.setModel(self.__model)
        self.setCompletionMode(completionMode)

        if parent:
            parent.setCompleter(self)


class DirCompleter(QCompleter):

    """Completer for directory names"""

    def __init__(self, parent=None,
                 completionMode=QCompleter.PopupCompletion,
                 showHidden=False):
        QCompleter.__init__(self, parent)
        self.__model = QDirModel(self)

        if showHidden:
            filters = QDir.Filters(QDir.Drives | QDir.AllDirs | QDir.Hidden)
        else:
            filters = QDir.Filters(QDir.Drives | QDir.AllDirs)
        self.__model.setFilter(filters)

        self.setModel(self.__model)
        self.setCompletionMode(completionMode)

        if parent:
            parent.setCompleter(self)


class StringListCompleter(QCompleter):

    """Completer for strings lists"""

    def __init__(self, parent=None, strings=None,
                 completionMode=QCompleter.PopupCompletion):
        if strings is None:
            strings = []
        QCompleter.__init__(self, parent)
        self.__model = QStringListModel(strings, parent)
        self.setModel(self.__model)
        self.setCompletionMode(completionMode)

        if parent:
            parent.setCompleter(self)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Provides the storage for the debugger environment"""

import os.path
from copy import deepcopy
from .fileutils import loadJSON, saveJSON


_DEFAULT_DEBUGGER_PROPS = {'breakpoints': [],   # [{'file': <str>,
                                                #   'line': <int>,
                                                #   'condition': <str>,
                                                #   'temp': <bool>,
                                                #   'enabled': <bool>,
                                                #   'ignorecnt': <int>}, ...]
                           'watchpoints': [],   # [{'condition': <str>,
                                                #   'special': <bool>,
                                                #   'temp': <bool>,
                                                #   'enabled': <bool>,
                                                #   'ignorecnt': <int>}, ...]
                           'ignoredexceptions': []}    # [type name, ... ]


class DebuggerEnvironment:

    """Loads/stores/saves the debugger environment"""

    def __init__(self):
        self.__props = deepcopy(_DEFAULT_DEBUGGER_PROPS)
        self.__deFileName = None

    def reset(self):
        """Resets the filesystem binding"""
        self.__props = deepcopy(_DEFAULT_DEBUGGER_PROPS)
        self.__deFileName = None

    def setup(self, dirName):
        """Binds the parameters to a disk file"""
        # Just in case - flush the previous data if they were bound
        DebuggerEnvironment.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for the debugger '
                            'environment. The given ' + dirName + ' is not.')

        self.__deFileName = dirName + "debuggerenv.json"
        if os.path.exists(self.__deFileName):
            DebuggerEnvironment.load(self)

    def load(self):
        """Loads the saved debugger environment"""
        if self.__deFileName:
            default = deepcopy(_DEFAULT_DEBUGGER_PROPS)
            self.__props = loadJSON(self.__deFileName, 'debugger environment',
                                    default)

    def save(self):
        """Saves the debugger environment into a file"""
        if self.__deFileName:
            saveJSON(self.__deFileName, self.__props, 'debugger environment')

    @property
    def breakpoints(self):
        """Provides the breakpoints"""
        return self.__props['breakpoints']

    @breakpoints.setter
    def breakpoints(self, bpointList):
        self.__props['breakpoints'] = bpointList
        DebuggerEnvironment.save(self)

    @property
    def watchpoints(self):
        """Provides the watchpoints"""
        return self.__props['watchpoints']

    @watchpoints.setter
    def watchpoints(self, wpointList):
        self.__props['watchpoints'] = wpointList
        DebuggerEnvironment.save(self)

    @property
    def exceptionFilters(self):
        """Provides the ignored exceptions"""
        return self.__props['ignoredexceptions']

    @exceptionFilters.setter
    def exceptionFilters(self, newFilters):
        self.__props['ignoredexceptions'] = newFilters
        DebuggerEnvironment.save(self)

    def addExceptionFilter(self, excptType):
        """Adds a new ignored exception type"""
        if excptType not in self.__props['ignoredexceptions']:
            self.__props['ignoredexceptions'].append(excptType)
            DebuggerEnvironment.save(self)

    def deleteExceptionFilter(self, excptType):
        """Remove ignored exception type"""
        if excptType in self.__props['ignoredexceptions']:
            self.__props['ignoredexceptions'].remove(excptType)
            DebuggerEnvironment.save(self)

    def addBreakpoint(self, fName, line):
        """Adds serialized breakpoint"""
        value = [fName, line]
        if value not in self.__props['breakpoints']:
            self.__props['breakpoints'].append(value)
            DebuggerEnvironment.save(self)

    def deleteBreakpoint(self, fName, line):
        """Deletes serialized breakpoint"""
        value = [fName, line]
        if value in self.__props['breakpoints']:
            self.__props['breakpoints'].remove(value)
            DebuggerEnvironment.save(self)

    def addWatchpoint(self, fName, expression):
        """Adds serialized watchpoint"""
        value = [fName, expression]
        if value not in self.__props['watchpoints']:
            self.__props['watchpoints'].append(value)
            DebuggerEnvironment.save(self)

    def deleteWatchpoint(self, fName, expression):
        """Deletes serialized watchpoint"""
        value = [fName, expression]
        if value in self.__props['watchpoints']:
            self.__props['watchpoints'].remove(value)
            DebuggerEnvironment.save(self)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2013 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Module implementing the Breakpoint model"""

from ui.qt import pyqtSignal, QAbstractItemModel, Qt, QModelIndex


COLUMN_LOCATION = 2
COLUMN_CONDITION = 3
COLUMN_TEMPORARY = 1
COLUMN_ENABLED = 0
COLUMN_IGNORE_COUNT = 4


class BreakPointModel(QAbstractItemModel):

    """Class implementing a custom model for breakpoints"""

    sigDataAboutToBeChanged = pyqtSignal(QModelIndex, QModelIndex)
    sigBreakpoinsChanged = pyqtSignal()

    def __init__(self, parent=None):
        QAbstractItemModel.__init__(self, parent)

        self.breakpoints = []

        self.__fields = {
            COLUMN_LOCATION: ['File:line', Qt.Alignment(Qt.AlignLeft)],
            COLUMN_CONDITION: ['Condition', Qt.Alignment(Qt.AlignLeft)],
            COLUMN_TEMPORARY: ['T', Qt.Alignment(Qt.AlignHCenter)],
            COLUMN_ENABLED: ['E', Qt.Alignment(Qt.AlignHCenter)],
            COLUMN_IGNORE_COUNT: ['Ignore Count', Qt.Alignment(Qt.AlignRight)]}
        self.__columnCount = len(self.__fields)

    def columnCount(self, parent=None):
        """Provides the current column count"""
        return self.__columnCount

    def rowCount(self, parent=None):
        """Provides the current row count"""
        # we do not have a tree, parent should always be invalid
        if parent is None or not parent.isValid():
            return len(self.breakpoints)
        return 0

    def data(self, index, role=Qt.DisplayRole):
        """Provides the requested data"""
        if not index.isValid():
            return None

        column = index.column()
        row = index.row()
        if role == Qt.DisplayRole:
            if column == COLUMN_LOCATION:
                return self.breakpoints[row].getLocation()
            if column == COLUMN_CONDITION:
                return self.breakpoints[row].getCondition()
            if column == COLUMN_IGNORE_COUNT:
                return self.breakpoints[row].getIgnoreCount()
        elif role == Qt.CheckStateRole:
            if column == COLUMN_TEMPORARY:
                return self.breakpoints[row].isTemporary()
            if column == COLUMN_ENABLED:
                return self.breakpoints[row].isEnabled()
        elif role == Qt.ToolTipRole:
            if column < self.__columnCount:
                return self.breakpoints[row].getTooltip()
        elif role == Qt.TextAlignmentRole:
            if column < self.__columnCount:
                return self.__fields[column][1]
        return None

    def setData(self, index, _, role=Qt.EditRole):
        """Change data in the model"""
        if index.isValid():
            if role == Qt.CheckStateRole:
                column = index.column()
                if column in [COLUMN_TEMPORARY, COLUMN_ENABLED]:
                    # Flip the boolean
                    row = index.row()
                    bp = self.breakpoints[row]

                    self.sigDataAboutToBeChanged.emit(index, index)
                    if column == COLUMN_TEMPORARY:
                        bp.setTemporary(not bp.isTemporary())
                    else:
                        bp.setEnabled(not bp.isEnabled())
                    self.dataChanged.emit(index, index)
                    self.sigBreakpoinsChanged.emit()
                    return True
        return False

    def flags(self, index):
        """Provides the item flags"""
        if not index.isValid():
            return Qt.ItemIsEnabled

        column = index.column()
        if column in [COLUMN_TEMPORARY, COLUMN_ENABLED]:
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsUserCheckable
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Provides header data"""
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section < self.__columnCount:
                return self.__fields[section][0]
            return ""
        return None

    def index(self, row, column, parent=None):
        """Creates an index"""
        if (parent and parent.isValid()) or \
           row < 0 or row >= len(self.breakpoints) or \
           column < 0 or column >= self.__columnCount:
            return QModelIndex()

        return self.createIndex(row, column, self.breakpoints[row])

    def parent(self, index):
        """Provides the parent index"""
        return QModelIndex()

    def hasChildren(self, parent=None):
        """Checks if there are child items"""
        if parent is None or not parent.isValid():
            return len(self.breakpoints) > 0
        return False

    def addBreakpoint(self, bpoint):
        """Adds a new breakpoint to the list"""
        cnt = len(self.breakpoints)
        self.beginInsertRows(QModelIndex(), cnt, cnt)
        self.breakpoints.append(bpoint)
        self.endInsertRows()
        self.sigBreakpoinsChanged.emit()

    def setBreakPointByIndex(self, index, bpoint):
        """Set the values of a breakpoint given by index"""
        if index.isValid():
            row = index.row()
            index1 = self.createIndex(row, 0, self.breakpoints[row])
            index2 = self.createIndex(row, self.__columnCount - 1,
                                      self.breakpoints[row])

            self.sigDataAboutToBeChanged.emit(index1, index2)
            self.breakpoints[row].update(bpoint)
            self.dataChanged.emit(index1, index2)
            self.sigBreakpoinsChanged.emit()

    def updateLineNumberByIndex(self, index, newLineNumber):
        """Update the line number by index"""
        if index.isValid():
            row = index.row()
            index1 = self.createIndex(row, 0, self.breakpoints[row])
            index2 = self.createIndex(row, self.__columnCount - 1,
                                      self.breakpoints[row])

            self.sigDataAboutToBeChanged.emit(index1, index2)
            self.breakpoints[row].updateLineNumber(newLineNumber)
            self.dataChanged.emit(index1, index2)
            self.sigBreakpoinsChanged.emit()

    def setBreakPointEnabledByIndex(self, index, enabled):
        """Sets the enable state"""
        if index.isValid():
            row = index.row()
            index1 = self.createIndex(row, 0, self.breakpoints[row])
            index2 = self.createIndex(row, self.__columnCount - 1,
                                      self.breakpoints[row])

            self.sigDataAboutToBeChanged.emit(index1, index2)
            self.breakpoints[row].setEnabled(enabled)
            self.dataChanged.emit(index1, index2)
            self.sigBreakpoinsChanged.emit()

    def deleteBreakPointByIndex(self, index):
        """Deletes the breakpoint by its index"""
        if index.isValid():
            row = index.row()
            self.beginRemoveRows(QModelIndex(), row, row)
            del self.breakpoints[row]
            self.endRemoveRows()
            self.sigBreakpoinsChanged.emit()

    def deleteBreakPoints(self, idxList):
        """Deletes a list of breakpoints"""
        rows = []
        for index in idxList:
            if index.isValid():
                rows.append(index.row())
        rows.sort(reverse=True)
        for row in rows:
            self.beginRemoveRows(QModelIndex(), row, row)
            del self.breakpoints[row]
            self.endRemoveRows()
        self.sigBreakpoinsChanged.emit()

    def deleteAll(self):
        """Deletes all breakpoints"""
        if self.breakpoints:
            self.beginRemoveRows(QModelIndex(), 0, len(self.breakpoints) - 1)
            self.breakpoints = []
            self.endRemoveRows()
            self.sigBreakpoinsChanged.emit()

    def getBreakPointByIndex(self, index):
        """Provides a breakpoint by index"""
        if index.isValid():
            return self.breakpoints[index.row()]
        return None

    def getBreakPointIndex(self, fname, lineno):
        """Provides an index of a breakpoint"""
        for row in range(len(self.breakpoints)):
            bpoint = self.breakpoints[row]
            if bpoint.getAbsoluteFileName() == fname and \
               bpoint.getLineNumber() == lineno:
                return self.createIndex(row, 0, self.breakpoints[row])
        return QModelIndex()

    def isBreakPointTemporaryByIndex(self, index):
        """Checks if a breakpoint given by it's index is temporary"""
        if index.isValid():
            return self.breakpoints[index.row()].isTemporary()
        return False

    def getCounts(self):
        """Provides enable/disable counters"""
        enableCount = 0
        disableCount = 0
        for bp in self.breakpoints:
            if bp.isEnabled():
                enableCount += 1
            else:
                disableCount += 1
        return enableCount, disableCount

    def serialize(self):
        """Provides a list of serialized breakpoints"""
        result = []
        for bp in self.breakpoints:
            result.append(bp.serialize())
        return result
